// DeloneTriangulation.cc

#include <sstream>
#include <cstdlib>

#include <cstring>

#include <cmath>
#include <gmp.h>

#include "Voronoi.h"
#include "DeloneTriangulation.h"
#include "la_support.h"
#include "lrslib.h"

SimplexContainer DeloneTriangulation::all_simplices;

//
//
//

void die_delone_triangulation(const char* last_words)
{
  std::cout <<"DeloneTriangulation.cc: " <<last_words;
  exit(EXIT_FAILURE);
}

//
//
//

DeloneTriangulation::DeloneTriangulation()
  : simplices(optimal_hash_table_size)
{
  no_neighbours = 0;
  max_volume=0;
  no_extreme=0;
  invar=-1;
}

DeloneTriangulation::DeloneTriangulation(const DeloneTriangulation& Del)
  : simplices(optimal_hash_table_size)
{
  simplices = Del.simplices;
  hash_simplices::iterator pos = simplices.begin();
  while (pos != simplices.end()) {
    all_simplices.add_simplex(*(*pos));
    pos++;
  }
  no_neighbours = 0;
  no_extreme=0;
  invar=Del.invar;
  max_volume=Del.max_volume;
}

//
//
//

DeloneTriangulation& DeloneTriangulation::operator=(const DeloneTriangulation& Del)
{
  if (this != &Del) {

    {
      destroy_neighbours();
      hash_simplices::iterator pos = simplices.begin();
      while (pos != simplices.end()) {
	all_simplices.delete_simplex(*(*pos));
	pos++;
      }
    }

    simplices = Del.simplices;
    hash_simplices::iterator pos = simplices.begin();
    while (pos != simplices.end()) {
      all_simplices.add_simplex(*(*pos));
      pos++;
    }
    no_neighbours = 0;
    no_extreme = 0;
	invar=Del.invar;
    max_volume=Del.max_volume;
  }
  return *this;
}

//
//
//

DeloneTriangulation::~DeloneTriangulation()
{
  destroy_neighbours();
  hash_simplices::iterator pos = simplices.begin();
  while (pos != simplices.end()) {
    all_simplices.delete_simplex(*(*pos));
    pos++;
  }
}

//
//
//

int DeloneTriangulation::compute_no_neighbours()
{
  if (no_neighbours == 0) 
    make_neighbours();
  return no_neighbours;
}

//
//
//

DeloneTriangulation DeloneTriangulation::compute_nth_neighbour(const int n)
{
  if (no_neighbours == 0)
    die_delone_triangulation("no_neighbours == 0.\n");
  if (!(n >= 0 && n <= no_neighbours))
    die_delone_triangulation("!(n >= 0 && n < no_neighbours).\n");
  return do_flip((*irredundants[n]).facets);
}

//
//
//

void DeloneTriangulation::read_COOP_file(FILE* coop_file)
{
  // dimension
  
  int dimension;
  fscanf(coop_file, "%d", &dimension);
  if (dim != dimension)
    die_delone_triangulation("dim != dimension.\n");

  // simplices
  
  int half_simplices_size;
  fscanf(coop_file, "%d", &half_simplices_size);

  double** vertices;
  doubleMakeSquareMatrix(dim+1, &vertices);

  for (int simplex_count = 0; simplex_count < half_simplices_size; simplex_count++) {
    for (int i = 0; i < dim+1; i++)
      vertices[0][i] = 0.0;
    for (int i = 0; i < dim+1; i++)
      vertices[i][dim] = 1.0;
    for (int i = 0; i < dim; i++)
      for (int j = 0; j < dim; j++)
	fscanf(coop_file, "%lf", &vertices[i+1][j]);

    Simplex L(vertices);
    Simplex* L1 = all_simplices.add_simplex(L);
    simplices.insert(L1);
    Simplex Lneg = L.compute_negative();
    Simplex* L2 = all_simplices.add_simplex(Lneg);
    simplices.insert(L2); 
  }

  doubleDestroySquareMatrix(dim+1, &vertices);

  // do not care about the rest
}

void DeloneTriangulation::write_COOP_file(FILE* coop_file) const
{
  // dimension
  fprintf(coop_file, "%d\n\n", dim);

  // simplices
  std::set<Simplex> half_simplices;
  hash_simplices::iterator simplices_pos = simplices.begin();
  while (simplices_pos != simplices.end()) {
    if (half_simplices.find(*(*simplices_pos)) == half_simplices.end() &&
	half_simplices.find((*(*simplices_pos)).compute_negative()) == half_simplices.end())
      half_simplices.insert(*(*simplices_pos));
    simplices_pos++;
  }
  fprintf(coop_file, "%d\n", half_simplices.size());
  for (std::set<Simplex>::iterator pos = half_simplices.begin(); pos != half_simplices.end(); pos++)
    (*pos).print_me(coop_file);

  // basis
  fprintf(coop_file, "%d\n", dim*(dim+1)/2);
  for (int i = 0; i < dim; i++)
    for (int j = 0; j <= i; j++)
      for (int row = 0; row < dim; row++) {
	for (int col = 0; col <= row; col++) 
	  if (row == i && col == j) fprintf(coop_file, "1 ");
	  else fprintf(coop_file, "0 ");
	fprintf(coop_file, "\n");
      }
  fprintf(coop_file, "\n");

  // inequality
  fprintf(coop_file, "%d\n", irredundants.size());
  for (unsigned int i = 0; i < irredundants.size(); i++) {
    for (int j = 0; j < dim*(dim+1)/2; j++)
      fprintf(coop_file, "%d ", int((*irredundants[i]).q[j]));
    fprintf(coop_file, "\n");
  }

  // max_iterations
  fprintf(coop_file, "\n500\n");

  // duality_gap
  fprintf(coop_file, "\n1e-10\n");
}

//
// compute_lower_bound
// see "Computational Approaches", Sec.10: Moments of Inertia
//

void DeloneTriangulation::compute_lower_bound(NTL::ZZ& invariant, double& lower_bound)
{
//  if(invar<0){
  	double** C;
    doubleMakeSquareMatrix(dim, &C);
    hash_simplices::iterator pos = simplices.begin();
    while (pos != simplices.end()) {
      (*(*pos)).inc_edge_matrix(C);
      pos++;
    }

  // compute invariant

    NTL::mat_ZZ zz_C;
    zz_C.SetDims(dim,dim);
    for (int i = 0; i < dim; i++)
      for (int j = 0; j < dim; j++)
        zz_C[i][j] = NTL::to_ZZ(C[i][j]);

    determinant(invariant, zz_C, true);

    invar=invariant;

  // compute lower_bound

  double q = double(simplices.size());
  double n = double(dim);
  double m = n/(q*(n+1)*(n+1));
  for (int i = 0; i < dim; i++)
    doubleMultiplyVector(dim, m, C[i], C[i]);
  
  determinant(dim, C, &lower_bound);    
  lower_bound = sqrt(fabs(lower_bound));

    doubleDestroySquareMatrix(dim, &C);
//  }
//  else{
//    invariant=invar;
//  }

}

void DeloneTriangulation::compute_invariant(NTL::ZZ& invariant)
{
  if(invar<0){
  	double** C;
    doubleMakeSquareMatrix(dim, &C);
    hash_simplices::iterator pos = simplices.begin();
    while (pos != simplices.end()) {
      (*(*pos)).inc_edge_matrix(C);
      pos++;
    }

  // compute invariant

    NTL::mat_ZZ zz_C;
    zz_C.SetDims(dim,dim);
    for (int i = 0; i < dim; i++)
      for (int j = 0; j < dim; j++)
        zz_C[i][j] = NTL::to_ZZ(C[i][j]);

    determinant(invariant, zz_C, true);

    invar=invariant;

    doubleDestroySquareMatrix(dim, &C);
  }
  else{
    invariant=invar;
  }

}


//
//
//

int permutation_sign(int *pi)
{
  int sign = 1;
  for (int i = 0; i < dim; i++)
    for (int j = i+1; j < dim; j++)
      if (pi[i] > pi[j]) sign = -sign;
  return sign;
}

//
//
//

bool check_equivalence(const hash_simplices& simplices0, const hash_simplices& simplices1, double** linear_map)
{
  bool equivalent = true;
  hash_simplices::iterator L_pos = simplices0.begin();
  Simplex L = *(*L_pos); // only to safe some constructors and destructors
  while (L_pos != simplices0.end() && equivalent) {
    L = (*(*L_pos)).perform_linear_map(linear_map);
    equivalent = (simplices1.find(&L) != simplices1.end());
    L_pos++;
  }
  return equivalent;
}

//
//
//

bool DeloneTriangulation::is_equivalent(DeloneTriangulation& Del) 
{
  double *form1;
  double *form2;

  doubleMakeVector(Dim, &form1);
  doubleMakeVector(Dim, &form2);

  if(Del.no_neighbours==0){ 
	Del.compute_no_neighbours();
  }
  if(no_neighbours==0) compute_no_neighbours();
  if (compute_no_extreme()!=Del.compute_no_extreme())
	return false;

  std::set<Extreme>::iterator ext_pos = extreme_forms.begin();

  while (ext_pos != extreme_forms.end()) {
    doubleAddVector(Dim, 1.0, (*ext_pos).q, form1);
    ext_pos++;
  }

  std::set<Extreme>::iterator ext_pos1 = Del.extreme_forms.begin();

  while (ext_pos1 != Del.extreme_forms.end()) {
    doubleAddVector(Dim, 1.0, (*ext_pos1).q, form2);
    ext_pos1++;
  }



  int temp;
  temp=-1;
  temp=isom(form1,form2);

  free(form1);
  free(form2);

  if(temp==0){
	return false;
  }
  else
	return true;
}

//
// 
//

DeloneTriangulation principal_domain()
{
  DeloneTriangulation Del;

  double** frame;
  doubleMakeSquareMatrix(dim+1, &frame);
  doubleSetIdentity(dim+1, frame);
  for (int i = 0; i < dim; i++)
    frame[dim][i] = -1.0;
  for (int i = 0; i < dim+1; i++)
    frame[i][dim] = 1.0;

  double** vertices;
  doubleMakeSquareMatrix(dim+1, &vertices);

  do {
    for (int i = 0; i < dim+1; i++)
      if (i != 0) {
	doubleCopyVector(dim+1, vertices[i-1], vertices[i]);
	doubleAddVector(dim, 1.0, frame[i], vertices[i]);
      }
      else
	doubleCopyVector(dim+1, frame[0], vertices[0]);
    Simplex L(vertices);
    // here we add too many simplices to all_simplices
    Simplex* L1 = &(const_cast<Simplex&>(*Del.all_simplices.add_simplex(L)));
    Del.simplices.insert(L1);
  }
  while (std::next_permutation(&(frame[0]), &(frame[dim+1])));

  doubleDestroySquareMatrix(dim+1, &vertices);
  doubleDestroySquareMatrix(dim+1, &frame);

  return Del;
}

//
//
//

DeloneTriangulation barnes_trenerry_domain()
{
  DeloneTriangulation Del;

  int *pi;
  intMakeVector(dim, &pi);
  for (int i = 0; i < dim; i++)
    pi[i] = i;

  double** frame;
  doubleMakeSquareMatrix(dim, &frame);
  doubleSetIdentity(dim, frame);

  double** vertices;
  doubleMakeSquareMatrix(dim+1, &vertices);

  int m = (dim-1)/2;

  do {
    
    for (int i = 0; i < dim+1; i++)
      for (int j = 0; j < dim; j++)
	vertices[i][j] = 0.0;
    for (int i = 0; i < dim+1; i++)
      vertices[i][dim] = 1.0;

    int vertex_pos = 1;

    for (int r = 1; r <= m; r++) {
      for (int i = 0; i < r; i++)
	doubleAddVector(dim, 1.0, frame[pi[i]], vertices[vertex_pos]);
      vertex_pos++;
    }
    for (int s = 0; s <= m; s++) {
      for (int i = 1; i <= dim; i++)
	if (i <= m+s)
	  doubleAddVector(dim, 0.5, frame[pi[i-1]], vertices[vertex_pos]);
	else
	  doubleAddVector(dim, -0.5, frame[pi[i-1]], vertices[vertex_pos]);
      vertex_pos++;
    }

    // apply transformation T
    for (int i = 1; i < dim+1; i++) {
      double y1 = vertices[i][0];
      vertices[i][0] = y1 + vertices[i][1];
      for (int j = 1; j < dim; j++)
	vertices[i][j] = y1 - vertices[i][j];
    }

    Simplex L(vertices);
    // here we add too many simplices to all_simplices
    Simplex* L1 = &(const_cast<Simplex&>(*Del.all_simplices.add_simplex(L)));
    Del.simplices.insert(L1);
  }
  while (std::next_permutation(&(pi[0]), &(pi[dim])));

  doubleDestroySquareMatrix(dim+1, &vertices);
  doubleDestroySquareMatrix(dim, &frame);
  intDestroyVector(dim, &pi);

  return Del;
}

//
//
//

void DeloneTriangulation::make_neighbours()
{
  if (no_neighbours != 0)
    die_delone_triangulation("second call of make_neighbours().\n");

  make_facets_and_regulators();

  // consistency check
  if (regulators.size() < unsigned(Dim))
    die_delone_triangulation("regulators.size() < unsigned(Dim).\n");

  // lrs


  if (!lrs_init(" * computing irredundant facets : "))
    die_delone_triangulation("lrs_init failed.\n");

  lrs_dat *Q;
  Q = lrs_alloc_dat("LRS globals");
  if (Q == NULL)
    die_delone_triangulation("lrs_alloc_dat failed.\n");
  Q->m = long(regulators.size());
  Q->n = long(Dim+1);
  Q->hull = TRUE;

  lrs_dic *P;
  P = lrs_alloc_dic(Q);
  if (P == NULL)
    die_delone_triangulation("lrs_alloc_dic failed.\n");

  // inequalities
  
  long num[Dim+1];
  long den[Dim+1];
  for (int i = 0; i < Dim+1; i++)
    den[i] = 1L;
  num[0] = 0L;
  
  long row = 1L;
  std::vector <std::set<Regulator>::iterator> indices(regulators.size());
  std::set<Regulator>::iterator reg_pos = regulators.begin();
  while (reg_pos != regulators.end()) {
    for (long i = 1; i < Dim+1; i++) 
      num[i] = long((*reg_pos).q[i-1]);
    lrs_set_row(P, Q, row, num, den, GE);
    indices[row-1] = reg_pos;
    
    row++;
    reg_pos++;
  }

  lrs_mp_matrix Lin;
  
//  itomp(ONE, Q->Lcm[0]); // circumvent a bug in lrs

  if (!lrs_getfirstbasis(&P, Q, &Lin, TRUE)) // TRUE means no_output
    die_delone_triangulation("lrs_getfirstbasis failed.\n");

  long m = P->m_A;
  long d = P->d;
  long nlinearity = Q->nlinearity;
  long lastdv = Q->lastdv;
  long *redineq = new long[m+1];
  long ineq;

  for (long i = 0; i < nlinearity; i++)
    redineq[Q->linearity[i]] = 2L;

  for (long index = lastdv + 1; index <= m + d; index++) {
    ineq = Q->inequality[index - lastdv];
    redineq[ineq] = checkindex (P, Q, index);
    if (redineq[ineq] != ONE) {
      irredundants.push_back(indices[ineq-1]);
    }
  }		   

  lrs_free_dic(P, Q);
  lrs_free_dat(Q);

  delete [] redineq;

  lrs_close(" * -> finished : ");

  // lrs end

  no_neighbours = irredundants.size();
}

//
//
//

void DeloneTriangulation::destroy_neighbours()
{
  if (no_neighbours != 0) {
    irredundants.clear();
    destroy_facets_and_regulators();
    no_neighbours = 0;
    no_extreme = 0;
  }
}

//
//
//

void DeloneTriangulation::make_facets_and_regulators()
{
  // make facets

  double** facet;
  doubleMakeSquareMatrix(dim+1, &facet);

  hash_simplices::iterator pos = simplices.begin();  
  int balance = 0;

  while (pos != simplices.end()) {
    for (int n = 0; n < dim+1; n++) {
      (*(*pos)).get_nth_facet(n, facet);
      std::pair<std::set<Facet>::iterator, bool> result = facets.insert(Facet(facet));
      if (!result.second) {
	const_cast<Facet&>(*result.first).insert_out_vertex(facet[dim]);
	balance--;
      }
      else
	balance++;
    }
    pos++;
  }
  doubleDestroySquareMatrix(dim+1, &facet);  

//  Made to work through 1354-th in dim 6.
//
//  if (balance != 0) 
//    die_delone_triangulation("balance != 0.\n");

  // make regulators

  std::set<Facet>::iterator facet_pos = facets.begin();
  while (facet_pos != facets.end()) {
    bool repartition = true;

    double **vertices;
    doubleMakeSquareMatrix(dim+1, &vertices);
    for (int i = 0; i < dim+1; i++) {
      if ((*facet_pos).alpha[i] < 0.0) {
	doubleCopySquareMatrix(dim+1, (*facet_pos).in_simplex, vertices);
	doubleCopyVector(dim+1, (*facet_pos).out_vertex, vertices[i]);
	Simplex Lnew(vertices);
	if (repartition) 
	  repartition = (simplices.find(&Lnew) != simplices.end());
      }
    }
    doubleDestroySquareMatrix(dim+1, &vertices);

    if (repartition) {
      std::pair<std::set<Regulator>::iterator, bool> result = regulators.insert(Regulator(*facet_pos));
      const_cast<Regulator&>(*result.first).insert_facet(&(*facet_pos));
    }
    facet_pos++;
  }
}

//
//
//

void DeloneTriangulation::destroy_facets_and_regulators()
{
  facets.clear();
  regulators.clear();
  extreme_forms.clear();
}

//
//
//

DeloneTriangulation DeloneTriangulation::do_flip(std::list<Facet*> flip_facets) const
{
  DeloneTriangulation Del;
  Del.simplices = simplices;
  hash_simplices::iterator pos = Del.simplices.begin();
  while (pos != Del.simplices.end()) {
    all_simplices.add_simplex(*(*pos));
    pos++;
  }

  double** vertices;
  doubleMakeSquareMatrix(dim+1, &vertices);

  while (!flip_facets.empty()) {
    Facet& facet = *flip_facets.front();

    // consistency check
    if (!(facet.alpha[dim] < 0.0))
      die_delone_triangulation("alpha[dim] >= 0.0.\n");
    
    for (int i = 0; i < dim+1; i++) {
      doubleCopySquareMatrix(dim+1, facet.in_simplex, vertices);
      doubleCopyVector(dim+1, facet.out_vertex, vertices[i]);

      Simplex L(vertices);
      if (facet.alpha[i] > 0.0) {
	// flip in vertex[i]
	Simplex* L1 = &(const_cast<Simplex&>(*Del.all_simplices.add_simplex(L)));
	Del.simplices.insert(L1);
      }
      if (facet.alpha[i] < 0.0) {
	// flip out vertex[i]
	Del.simplices.erase(&L);
      }
    }
    // flip out in_simplex
    Simplex L(facet.in_simplex);
    Del.simplices.erase(&L);
    flip_facets.pop_front();
  }

  doubleDestroySquareMatrix(dim+1, &vertices);

  // consistency check
  if (simplices.size() > unsigned(factorial(dim)))
    die_delone_triangulation("simplices.size() > factorial(dim).\n");
  
  return Del;
}

// A.G.

int DeloneTriangulation::no_simplices()
{
  return simplices.size();
}

// A.G.

double DeloneTriangulation::volume_sq_sum()
{
  hash_simplices::iterator pos = simplices.begin();  
  double sum = 0;
  

  while (pos != simplices.end()) {
	double tmp=(*(*pos)).compute_volume();
	sum = sum + tmp*tmp;
    pos++;
  }

  return sum;
}

int DeloneTriangulation::compute_max_volume(){
  if(max_volume!=0)
    return max_volume; 

  hash_simplices::iterator pos = simplices.begin();  
  
  while (pos != simplices.end()) {
	if( (*(*pos)).compute_volume()>max_volume)
      max_volume=(*(*pos)).compute_volume();
    pos++;
  }
  return max_volume;
}

Extreme DeloneTriangulation::compute_nth_extreme(const int n){
//  double* a;
//  doubleMakeVector(Dim,&a);
  if (no_extreme == 0)
    die_delone_triangulation("no_extreme == 0.\n");
  if (!(n >= 0 && n <= no_extreme))
    die_delone_triangulation("!(n >= 0 && n < no_extreme).\n");
  std::set<Extreme>::iterator ext_pos = extreme_forms.begin();
  
  for(int i=0;i<n;i++)
    ext_pos++;

  return *ext_pos;
}

int DeloneTriangulation::compute_no_extreme(){
  if(no_extreme==0){
	make_extreme();
  }
  return no_extreme;
}

void DeloneTriangulation::make_extreme(){
  if (no_neighbours == 0)
    die_delone_triangulation("no_neighbours == 0 while computing extreme forms.\n");

  if (no_extreme != 0)
    die_delone_triangulation("second call of make_extreme().\n");

//  std::vector<std::set<Regulator>::iterator>::iterator irr_pos = irredundants.begin();  

  double* temp;

  doubleMakeVector(Dim,&temp);

  lrs_mp_vector output;	/* one line of output:ray,vertex,facet,linearity */
  lrs_mp_matrix Lin;    /* holds input linearities if any are found      */


  long k,l;
  long col;    	/* output column index for dictionary            */
  
/*  for(unsigned int i=0; i<irredundants.size();i++){
    for(unsigned int j=0; j<Dim;j++)
	  temp[j]=(*irredundants[i]).q[j];
    printf("%lf\n",(*irredundants[i]).q[Dim-1]);
	extreme_forms.insert(Extreme(temp));
  }*/

  if (!lrs_init(" * computing extreme forms : "))
    die_delone_triangulation("lrs_init failed.\n");

  lrs_dat *Q;

  Q = lrs_alloc_dat ("LRS globals");
  
/* now flags in lrs_dat can be set */

  Q->n=long(Dim+1);           /* number of input columns         (dimension + 1 )  */
  Q->m=long(irredundants.size());         /* number of input rows = number of inequalities     */ 

/* manual alloc for output */

  output = (lrs_mp_vector)calloc ((Dim + 1), sizeof (lrs_mp ));

  for (k = 0; k <= Dim; k++)
    lrs_alloc_mp(output[k]);


  lrs_dic *P;
  P = lrs_alloc_dic(Q);
  if (P == NULL)
    die_delone_triangulation("lrs_alloc_dic failed.\n");

/* Build constraints of secondary cone */ 

long num[Dim+1];
long den[Dim+1];

num[0]=0;
den[0]=1;

for (k=1;k<=Q->m;k++){
  for(l=1;l<=Dim;l++){
    num [l] = (long)(*irredundants[k-1]).q[l-1];
    den [l] = 1;
  }
  lrs_set_row(P,Q,k,num,den,GE);
}


/* code from here is borrowed from lrs_main */

/* Pivot to a starting dictionary                      */

  if (!lrs_getfirstbasis (&P, Q, &Lin, TRUE))
    die_delone_triangulation("lrs_getfirstbasis failed.\n");

//  for (col = 0L; col < Q->nredundcol; col++)  /* print linearity space */
//    lrs_printoutput (Q, Lin[col]);      /* Array Lin[][] holds the coeffs.   */

/* We initiate reverse search from this dictionary       */
/* getting new dictionaries until the search is complete */
/* User can access each output line from output which is */
/* a vertex/ray/facet from the lrs_mp_vector output      */

  intMakeVector(dim,&extreme_ranks);

  Extreme current_form=Extreme(temp);

  do
    {
      for (col = 0; col <= P->d; col++)
        if (lrs_getsolution (P, Q, output, col))
          if (Q->hull || zero (output[0]))	/*non vertex */    {
            for (k = 0; k < Dim; k++){
          	  temp[k]=mpz_get_d(output[k+1]);
			}
			current_form=Extreme(temp);
			extreme_ranks[current_form.give_rank()-1]++;
			extreme_forms.insert(current_form);
		  }
    }
  while (lrs_getnextbasis (&P, Q, FALSE));

//  lrs_printtotals (P, Q);    /* print final totals */

/* free space : do not change order of next 3 lines! */

/* manual clear or output */

  for (k=0; k<=Dim; k++)
     lrs_clear_mp (output[k] );
  free (output);

  lrs_free_dic (P,Q);           /* deallocate lrs_dic */
  lrs_free_dat (Q);             /* deallocate lrs_dat */

 
 
  lrs_close (" * computing extreme forms :");



  no_extreme=extreme_forms.size();
/*  printf("no_extreme=%d\n",no_extreme);

  printf("Ranks of extreme forms:");
  for(k=0;k<dim;k++)
    printf(" %d",extreme_ranks[k]);
  printf("\n");*/

  doubleDestroyVector(Dim,&temp);
}

// codim - codimension of faces what we are looking for. If marker==(1..1) then codim=1.

int DeloneTriangulation::subset_facets(const int codim, const int* marker, int** res_marker){
	int i,j,k;
	int quan=0;
	int extr=compute_no_extreme();
	for(i=0;i<extr;i++)
		quan+=marker[i];
	if(quan==(Dim-codim+1)){
		(*res_marker)=(int*)malloc((quan*extr)*sizeof(int));
		k=0;
		for(i=0;i<quan;i++){
			while(marker[k]!=1){
				k++;
			}
			for(j=0;j<extr;j++){
				(*res_marker)[i*extr+j]=marker[j];
			}
			(*res_marker)[i*extr+k]=0;
			k++;
		}
		return quan;
	}
	else{
		int quan=0;
		int max_quan=irredundants.size();
		int* temp_marker;
		temp_marker=(int*)malloc(max_quan*extr*sizeof(int));
		double** temp_mat;
		temp_mat=(double**)malloc(extr*sizeof(double*));
		int cur_num=0;
		int* cur_facet;
		cur_facet=(int*)malloc(extr*sizeof(int));
		for(int i=0;i<max_quan;i++){
			cur_num=0;
			for(int j=0;j<extr;j++){
				if(marker[j]==1){
					double value=0;
					for(int k=0;k<Dim;k++){
						value+=(((*irredundants[i]).q[k])*(compute_nth_extreme(j).q[k]));
					}
					if((value<0.001)&&(value>-0.001)){
						cur_facet[cur_num]=j;
						temp_mat[cur_num]=(double*)malloc(Dim*sizeof(double));
						for(int k=0;k<Dim;k++)
							temp_mat[cur_num][k]=compute_nth_extreme(j).q[k];
						cur_num++;					
					}
				}
			}
			int rank=0;

			if(cur_num>=(Dim-codim)){

				double mult,row_norm;
				int ii=0;

				while(ii<cur_num){
					row_norm=0.0;
		
					for(int jj=0;jj<Dim;jj++)
						row_norm=row_norm+temp_mat[ii][jj]*temp_mat[ii][jj];
		
					if(row_norm>0.0001){
						int jj=0;
						while((temp_mat[ii][jj]<0.001)&&(temp_mat[ii][jj]>-0.001))
							jj++;
						for(int kk=ii+1;kk<cur_num;kk++){
							mult=-temp_mat[kk][jj]/temp_mat[ii][jj];

							for(int ll=0;ll<Dim;ll++)
								temp_mat[kk][ll]=temp_mat[kk][ll]+mult*temp_mat[ii][ll];
						}
						rank++;
					}
					ii++;
				}
			}

			if(rank==(Dim-codim)){
				bool found=false;

				for(int ii=0; (ii<quan)&&(!found); ii++){
					found=true;
					for(int jj=0;(jj<cur_num)&&found;jj++)
						if(temp_marker[ii*extr+cur_facet[jj]]==0)
							found=false;
				}	
				if(!found){
					for(int j=0;j<extr;j++){
						temp_marker[quan*extr+j]=0;
					}
					for(int j=0;j<cur_num;j++){
						temp_marker[quan*extr+cur_facet[j]]=1;
					}
					quan++;
				}
			}
			for(int i=0;i<cur_num;i++)
				free(temp_mat[i]);
		}
		(*res_marker)=(int*)malloc((quan*extr)*sizeof(int));
		for(int i=0;i<quan*extr;i++){
			(*res_marker)[i]=temp_marker[i];
		}

		free(cur_facet);
		free(temp_mat);
		free(temp_marker);
		return quan;
	}	
}

void DeloneTriangulation::subset(const int *mark, QForm* res)
{
//  QForm temp;
//  (*res).codim=-1;
  int rank;
/*  if(compute_no_extreme()==Dim)
  {
	(*res).codim=Dim-quan;
  }*/
//  intMakeVector(dim,&temp.ranks);
//  doubleMakeVector(Dim,&temp.q);

  for(int i=0;i<dim;i++)
	(*res).ranks[i]=0;

  for(int i=0;i<Dim;i++)
	(*res).q[i]=0;

  for(int i=0;i<no_extreme;i++)
  { 
	if(mark[i]!=0){
	  rank=compute_nth_extreme(i).give_rank();
	  (*res).ranks[rank-1]++;
	  doubleAddVector(Dim,1.0,compute_nth_extreme(i).q,(*res).q);
	}
  }
//  return temp;
}

void DeloneTriangulation::write_form_file(QForm form, FILE* form_file)
{
	fprintf(form_file,"dimension: %d\n\nform:\n",dim);
	for(int i=0;i<dim;i++){
		for(int j=0; j<=i;j++){
			fprintf(form_file,"%4d",(int)form.q[i*(i+1)/2+j]);
		}
		fprintf(form_file,"\n");
	}
	fprintf(form_file,"\n");
	int quan=0;
	for(int i=0;i<no_extreme;i++)
		if(form.marker[i]==1)
			quan++;
	fprintf(form_file,"codimension: %d\n\nnumber of extremal forms: %d\n\n",form.codim,quan);

	fprintf(form_file,"number of extremal forms of each rank:");
	for(int i=0;i<dim;i++)
		fprintf(form_file," %d",form.ranks[i]);
	fprintf(form_file,"\n\n");

	fprintf(form_file,"extremal forms\n\n");

	for(int i=0;i<no_extreme;i++)
		if(form.marker[i]==1){
			fprintf(form_file,"rank: %d\nform:\n",compute_nth_extreme(i).give_rank());
			for(int j=0;j<dim;j++){
				for(int k=0; k<=j;k++){
					fprintf(form_file,"%4d",(int)compute_nth_extreme(i).q[j*(j+1)/2+k]);
				}
				fprintf(form_file,"\n");
			}
			fprintf(form_file,"\n");
		}
}

/* functions for isom */

#ifdef ARITH64
void single2Print(FILE *f, TheNumeric TheNb)
{
  fprintf(f, "%2lld", TheNb);
}
void singleScan(FILE *f, TheNumeric *TheNb)
{
  fscanf(f, "%lld", TheNb);
}
TheNumeric GetMAXNORM()
{
  TheNumeric a;
  int i;
  a=1;
  for (i=1; i<=18; i++)
    {
      a=a*10;
    }
  return a;
}
TheNumeric GetMAXENTRY()
{
  TheNumeric a;
  int i;
  a=1;
  for (i=1; i<=8; i++)
    {
      a=a*10;
    }
  return a;
}
#else
void single2Print(FILE *f, TheNumeric TheNb)
{
  fprintf(f, "%2d", TheNb);
}
void singleScan(FILE *f, TheNumeric *TheNb)
{
  fscanf(f, "%d", TheNb);
}
TheNumeric GetMAXNORM()
{
  TheNumeric a;
  int i;
  a=1;
  for (i=1; i<=8; i++)
    {
      a=a*10;
    }
  return a;
}
TheNumeric GetMAXENTRY()
{
  TheNumeric a;
  int i;
  a=1;
  for (i=1; i<=4; i++)
    {
      a=a*10;
    }
  return a;
}
#endif

void putform(TheNumeric **A, int n)
/*****	prints nxn-matrix A with format conventions as 
      getmat()	*****/
{
	int	i, j, symm, skew, diag, scal;

	symm = 1;
	skew = 1;
	diag = 1;
	scal = 1;
	for (i = 0; i < n  &&  (symm == 1  ||  skew == 1); ++i)
	{
		for (j = 0; j <= i  &&  (symm == 1  ||  skew == 1); ++j)
		{
			if (A[i][j] != A[j][i])
/* the matrix is not symmetric and hence not diagonal or scalar */
			{
				symm = 0;	
				diag = 0;
				scal = 0;
			}
			if (A[i][j] != -A[j][i])
/* the matrix is not skew-symmetric */
				skew = 0;	
			if (j != i  &&  A[i][j] != 0)
/* the matrix is not diagonal and hence not scalar */
			{
				diag = 0;
				scal = 0;
			}
		}
		if (i > 0  &&  A[i][i] != A[i-1][i-1])
/* the matrix is not scalar */
			scal = 0;
	}
	if (scal == 1)
		printf("%dd0\n", n);
	else if (diag == 1)
		printf("%dd1\n", n);
	else if (symm == 1)
		printf("%dx0\n", n);
	else if (skew == 1)
		printf("%dx-1\n", n);
	else
		printf("%d\n", n);
	for (i = 0; i < n; ++i)
	{
		if (scal == 1)
		{
			if (i == 0)
			  {
			    printf(" ");
			    single2Print(stdout, A[0][0]);
			    // printf("%2d", A[0][0]);
			    printf("\n");
			  }
		}
		else if (diag == 1)
		{
		  printf(" ");
		  single2Print(stdout, A[i][i]);
		  if (i == n-1)
		    printf("\n");
		}
		else if (symm == 1  ||  skew == 1)
		{
			for (j = 0; j <= i; ++j)
			  {
			    printf(" ");
			    single2Print(stdout, A[i][j]);
			  }
			printf("\n");
		}
		else
		{
			for (j = 0; j < n; ++j)
			  {
			    printf(" ");
			    single2Print(stdout, A[i][j]);
			  }
			printf("\n");
		}
	}
}



void putiso(TheNumeric **X, int dim1)
/*****	prints an isometry in ASCII-, GAP- or
      MAGMA-format	*****/
{
	int	i, j;

	printf("%d\n", dim1);
	
	for (i = 0; i < dim1; ++i)
	{
		for (j = 0; j < dim1-1; ++j)
		{
			    printf(" ");
			    single2Print(stdout, X[i][j]);
		}
		    printf(" ");
		    single2Print(stdout, X[i][dim1-1]);
		    printf("\n");
	}
}

int isocand(int *CI, int I, int *x, veclist V, invar F, invar FF, fpstruct fp)
/*****	tests, whether x[0],...,x[I-1] is a partial 
      isometry, using scalar product combinations
      and Bacher-polynomials depending on the chosen 
      options, puts the candidates for x[I] (i.e. the
      vectors vec such that the scalar products of 
      (x[0],...,x[I-1],vec) are correct) on CI, 
      returns their number 
      (should be fp.diag[I])	*****/
{
	int	i, j, k, dim1, okp, okm, sign, nr, fail, num;
	int	DEP, len, rank, n, xk;
	TheNumeric *FAiI, *Vvj, **Fvi, *comtri;
	TheNumeric **xbase, **Fxbase, *scpvec;
	TheNumeric *xnum, **xvec, *vec, vj, xbij, *comcoi;
	scpcomb	com;

	dim1 = F.dim1;
	DEP = 0;
	len = 0;

	if ((vec = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
		exit (1);
	if ((scpvec = (TheNumeric*)malloc(len * sizeof(TheNumeric))) == 0)
		exit (1);

/* CI is the list for the candidates */
	for (i = 0; i < fp.diag[I]; ++i)
		CI[i] = 0;
	nr = 0;
	fail = 0;
	for (j = 1; j <= V.n  &&  fail == 0; ++j)
	{
		Vvj = V.v[j];
		okp = 0;
		okm = 0;
		for (k = 0; k < len; ++k)
			scpvec[k] = 0;
		for (i = 0; i < 1; ++i)
		{
			FAiI = F.A[i][fp.per[I]];
			Fvi = FF.v[i];
/* vec is the vector of scalar products of V.v[j] with the first I base vectors
   x[0]...x[I-1] with respect to the forms in FF */
			for (k = 0; k < I; ++k)
			{
				if ((xk = x[k]) > 0)
					vec[k] = sscp(Vvj, Fvi[xk], dim1);
				else
					vec[k] = -sscp(Vvj, Fvi[-xk], dim1);
			}
			for (k = 0; k < I  &&  vec[k] == FAiI[fp.per[k]]; ++k);
			if (k == I  &&  Vvj[dim1+i] == FAiI[fp.per[I]])
/* V.v[j] is a candidate for x[I] with respect to the form F.A[i] */
				++okp;
			for (k = 0; k < I  &&  vec[k] == -FAiI[fp.per[k]]; ++k);
			if (k == I  &&  Vvj[dim1+i] == FAiI[fp.per[I]])
/* -V.v[j] is a candidate for x[I] with respect to the form F.A[i] */
				++okm;
		}
		if (okp == 1)
/* V.v[j] is a candidate for x[I] */
		{
			if (nr < fp.diag[I])
			{
				CI[nr] = j;
				++nr;
			}
			else
/* there are too many candidates */
				fail = 1;
		}
		if (okm == 1)
/* -V.v[j] is a candidate for x[I] */
		{
			if (nr < fp.diag[I])
			{
				CI[nr] = -j;
				++nr;
			}
			else
/* there are too many candidates */
				fail = 1;
		}
	}
	if (fail == 1)
		nr = 0;
	if (nr == fp.diag[I]  &&  I-1 >= 0  &&  DEP > 0)
/* compute the basis of the lattice generated by the vectors in xvec via the
   transformation matrix comb[I-1].trans */
	{
		for (i = 0; i < rank; ++i)
		{
			comtri = com.trans[i];
			for (j = 0; j < dim1; ++j)
			{
				xbij = 0;
				for (k = 0; k <= n; ++k)
					xbij += comtri[k] * xvec[k][j];
				xbase[i][j] = xbij;
			}
		}
	}
	if (nr == fp.diag[I]  &&  I-1 >= 0  &&  DEP > 0)
/* check, whether the base xbase has the right scalar products */
	{
		for (i = 0; i < 1  &&  nr > 0; ++i)
		{
			for (j = 0; j < rank; ++j)
			{
				for (k = 0; k < dim1; ++k)
					Fxbase[j][k] = sscp(FF.A[i][k], xbase[j], dim1);
			}
			for (j = 0; j < rank  &&  nr > 0; ++j)
			{
				for (k = 0; k <= j  &&  nr > 0; ++k)
				{
					if (sscp(xbase[j], Fxbase[k], dim1) != com.F[i][j][k])
/* a scalar product is wrong */
						nr = 0;
				}
			}
		}
	}
	free(scpvec);
	free(vec);
	return(nr);
}

int isometry(veclist V, invar F, invar FF, fpstruct fp, TheNumeric ***G, int nG)
/*****	search for an isometry	*****/
{
        int	i, dim1, **C, found, *x;
	TheNumeric **X;
	dim1 = F.dim1;
	if ((C = (int**)malloc(dim1 * sizeof(int*))) == 0)
		exit (1);
/* C[i] is the list of candidates for the image of the i-th base-vector */
	for (i = 0; i < dim1; ++i)
	{
		if ((C[i] = (int*)malloc(fp.diag[i] * sizeof(int))) == 0)
			exit (1);
	}
/* x is the new base, i.e. x[i] is the index in V.v of the i-th base-vector */
	if ((x = (int*)malloc(dim1 * sizeof(int))) == 0)
		exit (1);
/* compute the candidates for x[0] */
	isocand(C[0], 0, x, V, F, FF, fp);
	found = 0;
/* go into the recursion */
	found = iso(0, x, C, V, F, FF, fp, G, nG);
/*	if (found == 1)
	{
		for (i = 0; i < dim1; ++i)
			free(C[i]);
		free(C);
		free(x);

		return 1;
	}*/
	for (i = 0; i < dim1; ++i)
		free(C[i]);
	free(C);
	free(x);
	return found;
}

int iso(int step, int *x, int **C, veclist V, invar F, invar FF, fpstruct fp, TheNumeric ***G, int nG)
/*****	the heart of the program: the recursion	*****/
{
	FILE	*outfile;
	int	i, j, dim1, nH, *orb, orblen, found, Maxfail;
	TheNumeric ***H;
	dim1 = F.dim1;
	found = 0;
	while (C[step][0] != 0  &&  found == 0)
	{
		if (step < dim1-1)
		{
/* choose the image of the base-vector nr. step */
			x[step] = C[step][0];
/* check, whether x[0]...x[step] is a partial automorphism and compute the
   candidates for x[step+1] */
			if (isocand(C[step+1], step+1, x, V, F, FF, fp) == fp.diag[step+1])
			{
/* go deeper into the recursion */
				Maxfail = 0;
/* determine the heuristic value of Maxfail for the break condition in 
   isostab */
				for (i = 0; i <= step; ++i)
				{
					if (fp.diag[i] > 1)
						Maxfail += 1;
				}
				for (i = step+1; i < dim1; ++i)
				{
					if (fp.diag[i] > 1)
						Maxfail += 2;
				}
/* compute the stabilizer H of x[step] in G */
				nH = isostab(&H, x[step], G, nG, V, Maxfail);
				found = iso(step+1, x, C, V, F, FF, fp, H, nH);
			}
			if (found == 1)
			{
				for (i = 0; i < nG; ++i)
				{
					for (j = 0; j < dim1; ++j)
						free(G[i][j]);
					free(G[i]);
				}
				free(G);
				return(found);
			}
			orblen = orbit(&(x[step]), 1, G, nG, V, &orb);
/* delete the orbit of the chosen vector from the list of candidates */
			delete_new(C[step], fp.diag[step], orb, orblen);
			free(orb);
		}
		else
		{
/* an isomorphism is found */
			x[dim1-1] = C[dim1-1][0];
			found = 1;
		}
	}
	for (i = 0; i < nG; ++i)
	{
		for (j = 0; j < dim1; ++j)
			free(G[i][j]);
		free(G[i]);
	}
	free(G);
	return(found);
}

int isostab(TheNumeric ****H, int pt, TheNumeric ***G, int nG, veclist V, int Maxfail)
/*****	computes the orbit of V.v[pt] 
      under the generators 
      G[0],...,G[nG-1] and elements
      stabilizing V.v[pt], which are 
      stored in H, returns the number
      of generators in H	*****/
{
	int	*orb, len, cnd, orblen, tmplen, rpt;
	int	*flag;
	int	i, j, k, dim1, im, nH, fail;
	TheNumeric ***w, **A, **B;

/* a heuristic break condition for the computation of stabilizer elements:
   it would be too expensive to calculate all the stabilizer generators, which
   are obtained from the orbit, since this is highly redundant, 
   on the other hand every new generator which enlarges the group reduces the
   number of orbits and hence the number of candidates to be tested,
   after Maxfail subsequent stabilizer elements, that do not enlarge the group,
   the procedure stops,
   increasing Maxfail will possibly decrease the number of tests,
   but will increase the running time of the stabilizer computation
   there is no magic behind the heuristic, tuning might be appropriate */
	dim1 = V.dim1;
/* H are the generators of the stabilizer of V.v[pt] */
	if ((*H = (TheNumeric***)malloc(1 * sizeof(TheNumeric**))) == 0)
		exit (1);
	if (((*H)[0] = (TheNumeric**)malloc(dim1 * sizeof(TheNumeric*))) == 0)
		exit (1);
	for (i = 0; i < dim1; ++i)
	{
		if (((*H)[0][i] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
			exit (1);
	}
	nH = 0;
/* w[i+V.n] is a matrix that maps V.v[pt] on V.v[i] */
	if ((w = (TheNumeric***)malloc((2*V.n+1) * sizeof(TheNumeric**))) == 0)
		exit (1);
/* orb contains the orbit of V.v[pt] */
	if ((orb = (int*)malloc(2*V.n * sizeof(int))) == 0)
		exit (1);
/* orblen is the length of the orbit of a random vector in V.v */
	orblen = 1;
	for (i = 0; i < 2*V.n; ++i)
		orb[i] = 0;
/* if flag[i+V.n] = 1, then the point i is already in the orbit */
	if ((flag = (int*)malloc((2*V.n+1) * sizeof(int))) == 0)
		exit (1);
	for (i = 0; i <= 2*V.n; ++i)
		flag[i] = 0;
	orb[0] = pt;
	flag[orb[0]+V.n] = 1;
/* w[pt+V.n] is the Identity */
	if ((w[orb[0]+V.n] = (TheNumeric**)malloc(dim1 * sizeof(TheNumeric*))) == 0)
		exit (1);
	for (i = 0; i < dim1; ++i)
	{
		if ((w[orb[0]+V.n][i] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
			exit (1);
		for (j = 0; j < dim1; ++j)
			w[orb[0]+V.n][i][j] = 0;
		w[orb[0]+V.n][i][i] = 1;
	}
/* A and B are matrices for temporary use */
	if ((A = (TheNumeric**)malloc(dim1 * sizeof(TheNumeric*))) == 0)
		exit (1);
	for (i = 0; i < dim1; ++i)
	{
		if ((A[i] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
			exit (1);
	}
	if ((B = (TheNumeric**)malloc(dim1 * sizeof(TheNumeric*))) == 0)
		exit (1);
	for (i = 0; i < dim1; ++i)
	{
		if ((B[i] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
			exit (1);
	}
	cnd = 0;
	len = 1;
/* fail is the number of successive failures */
	fail = 0;
	while (len > cnd  &&  fail < Maxfail)
	{
		for (i = 0; i < nG  &&  fail < Maxfail; ++i)
		{
			im = operate(orb[cnd], G[i], V);
			if (flag[im+V.n] == 0)
/* a new element is found, appended to the orbit and an element mapping
   V.v[pt] to im is stored in w[im+V.n] */
			{
				orb[len] = im;
				flag[im+V.n] = 1;
				if ((w[im+V.n] = (TheNumeric**)malloc(dim1 * sizeof(TheNumeric*))) == 0)
					exit (1);
				for (j = 0; j < dim1; ++j)
				{
					if ((w[im+V.n][j] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
						exit (1);
				}
				matmul(w[orb[cnd]+V.n], G[i], dim1, w[im+V.n]);
				++len;
			}
			else
/* the image was already in the orbit */
			{
				matmul(w[orb[cnd]+V.n], G[i], dim1, B);
/* check, whether the old and the new element mapping pt on im differ */
				for (j = 0; j < dim1  &&  comp(B[j], w[im+V.n][j], dim1) == 0; ++j);
				if (j < dim1)
				{
/* w[im+V.n] is copied to A, since psolve changes the matrices */
					for (j = 0; j < dim1; ++j)
					{
						for (k = 0; k < dim1; ++k)
							A[j][k] = w[im+V.n][j][k];
					}
/* new stabilizer element H[nH] = w[orb[cnd]+V.n] * G[i] * (w[im+V.n])^-1 */
					psolve((*H)[nH], A, B, dim1, V.prime);
					rpt = rand() % V.n + 1;
					tmplen = orbitlen(rpt, 2*V.n, *H, nH+1, V);
					while (tmplen < orblen)
/* the orbit of this vector is shorter than a previous one, hence choose a new
   random vector */
					{
						rpt = rand() % V.n + 1;
						tmplen = orbitlen(rpt, 2*V.n, *H, nH+1, V);
					}
					if (tmplen > orblen)
/* the new stabilizer element H[nH] enlarges the group generated by H */
					{
						orblen = tmplen;
						++nH;
/* allocate memory for the new generator */
						if ((*H = (TheNumeric***)realloc(*H, (nH+1) * sizeof(TheNumeric**))) == 0)
							exit (1);
						if (((*H)[nH] = (TheNumeric**)malloc(dim1 * sizeof(TheNumeric*))) == 0)
							exit (1);
						for (k = 0; k < dim1; ++k)
						{
							if (((*H)[nH][k] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
								exit (1);
						}
						fail = 0;
					}
					else
/* the new stabilizer element does not enlarge the orbit of a random vector */
						++fail;
				}
/* if H[nH] is the identity, nothing is done */
			}
		}
		++cnd;
	}
	for (i = 0; i <= 2*V.n; ++i)
	{
		if (flag[i] == 1)
		{
			for (j = 0; j < dim1; ++j)
				free(w[i][j]);
			free(w[i]);
		}
	}
	free(w);
	for (i = 0; i < dim1; ++i)
	{
		free(A[i]);
		free(B[i]);
		free((*H)[nH][i]);
	}
	free(A);
	free(B);
	free((*H)[nH]);
	free(orb);
	free(flag);
	return(nH);
}

int lll(TheNumeric **F, TheNumeric **T, int dim1)
/*****	LLL-reduction of the positive semidefinite 
      Gram-matrix F with transformation matrix T	*****/
{
	int	i, j, m, l, rank, tmp;
	double	**mu, *B;
	TheNumeric **gram, *ptmp;

/* the weights */
	if ((B = (double*)malloc(dim1 * sizeof(double))) == 0)
		exit (1);
/* the model matrix */
	if ((mu = (double**)malloc(dim1 * sizeof(double*))) == 0)
		exit (1);
/* F is copied on gram and remains unchanged */
	if ((gram = (TheNumeric**)malloc(dim1 * sizeof(TheNumeric*))) == 0)
		exit (1);
	for (i = 0; i < dim1; ++i)
	{
		if ((mu[i] = (double*)malloc(dim1 * sizeof(double))) == 0)
			exit (1);
		if ((gram[i] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
			exit (1);
		for (j = 0; j < dim1; ++j)
		{
			T[i][j] = 0;
			mu[i][j] = 0.0;
			gram[i][j] = F[i][j];
		}
		T[i][i] = 1;
		mu[i][i] = 1.0;
	}
/* the first basis-vector should not have norm 0 */
	for (i = 0; i < dim1  &&  gram[i][i] == 0; ++i);
	if (i == dim1)
		return(0);
	else if (i > 0)
	{
		ptmp = gram[i];
		gram[i] = gram[0];
		gram[0] = ptmp;
		for (j = 0; j < dim1; ++j)
		{
			tmp = gram[j][i];
			gram[j][i] = gram[j][0];
			gram[j][0] = tmp;
		}
		T[0][0] = 0;
		T[0][i] = 1;
		T[i][i] = 0;
		T[i][0] = 1;
	}
	rank = dim1;
	initialize(0, mu, B, gram);
	if (dim1 > 1)
		initialize(1, mu, B, gram);
/* recursively go through the matrix */
	m = 1;
	l = 0;
	while (m < rank)
	  m = red(&m, &l, gram, T, mu, B, dim1, &rank);
	free(B);
	for (i = 0; i < dim1; ++i)
	{
		free(mu[i]);
		free(gram[i]);
	}
	free(mu);
	free(gram);
	return(rank);
}

void initialize(int i, double **mu, double *B, TheNumeric **gram)
/*****	initialization of the model	*****/
{
  int	j, k;
  double	muh, *mui, *muj;
  
  B[i] = (double)gram[i][i];
  mui = mu[i];
  for (j = 0; j < i; ++j)
    {
      muh = (double)gram[i][j];
      muj = mu[j];
      for (k = 0; k < j; ++k)
	muh -= B[k] * mui[k] * muj[k];
      muh /= B[j];
      B[i] -= B[j] * muh * muh;
      mui[j] = muh;
    }
}

int red(int *m, int *l, TheNumeric **gram, TheNumeric **T, double **mu, double *B, int dim1, int *rank)
/*****	pair-reduction with vectors 
      m and l, changes Gram-matrix, 
      transformation matrix and 
      model appropriately	*****/
{
	double	r;
	int	j, ir, jump;
	TheNumeric *gramm, *graml;
	TheNumeric *Tm, *Tl, *ptmp;
	jump = 0;
	r = mu[*m][*l];
	if (r > 0.5  ||  r < -0.5)
	{
		ir = MyRound(r);
		Tm = T[*m];
		Tl = T[*l];
		gramm = gram[*m];
		graml = gram[*l];
		for (j = 0; j < dim1; ++j)
		{
			Tm[j] -= ir * Tl[j];
			gramm[j] -= ir * graml[j];	
		}
		for (j = 0; j < dim1; ++j)
			gram[j][*m] -= ir * gram[j][*l];	
		initialize(*m, mu, B, gram); 
	}
	gramm = gram[*m];
	for (j = 0; j < *rank  &&  gramm[j] == 0; ++j);
	if (j == *rank)
	{
		ptmp = T[*m];
		T[*m] = T[*rank-1];
		T[*rank-1] = ptmp;
		for (j = 0; j < dim1; ++j)
		{
			gram[*m][j] = gram[*rank-1][j];
			gram[*rank-1][j] = 0;
		}
		for (j = 0; j < dim1; ++j)
		{
			gram[j][*m] = gram[j][*rank-1];
			gram[j][*rank-1] = 0;
		}
		*rank -= 1;
		if (*m < *rank)
		{
			initialize(*m, mu, B, gram); 
			*l = *m-1;
			jump = 1;
		}
		else
			jump = 1;
	}
	if (*l == *m-1  &&  jump == 0)
		check(B, m, l, gram, T, mu, dim1, rank);
	else if (jump == 0)
		decrease(m, l, gram, mu, B, rank);
	return(*m);
}

void check(double *B, int *m, int *l, TheNumeric **gram, TheNumeric **T, double **mu, int dim1, int *rank)
/*****	check the 
      LLL-condition	*****/
{
  if (B[*m] < (LLL_CONST - mu[*m][*m-1]*mu[*m][*m-1]) * B[*m-1])
    interchange(m, l, gram, T, mu, B, dim1);
  else
    {
      *l = *m-1;
      decrease(m, l, gram, mu, B, rank);	
    }
}

void decrease(int *m, int *l, TheNumeric **gram, double **mu, double *B, int *rank)
/*****	go back in the recursion 
      and change the model	*****/
{
  *l -= 1;
  if (*l < 0)
    {
      *m += 1;
      if (*m < *rank)
	{
	  *l = *m-1;
	  initialize(*m, mu, B, gram);
	}
    }
}

void interchange(int *m, int *l, TheNumeric **gram, TheNumeric **T, double **mu, double *B, int dim1)
/*****	interchange the vectors 
      nr. m and m-1 and 
      change the model 
      appropriately	*****/
{
	int	i, tmp;
	TheNumeric *ptmp;
	ptmp = T[*m];
	T[*m] = T[*m-1];
	T[*m-1] = ptmp;
	ptmp = gram[*m];
	gram[*m] = gram[*m-1];
	gram[*m-1] = ptmp;
	for (i = 0; i < dim1; ++i)
	{
		tmp = gram[i][*m];
		gram[i][*m] = gram[i][*m-1];
		gram[i][*m-1] = tmp;
	}
	initialize(*m-1, mu, B, gram);
	if (*m > 1)
	{
		*m -= 1;
		*l = *m-1;
	}
	else
	{
		initialize(*m, mu, B, gram);
		*l = *m-1;
	}
}

TheNumeric MyRound(double x)
/*****	rounds x to an integer	*****/
{
	if (x >= 0)
	  {
	    return((TheNumeric)(2.0*x + 1.0) / 2);
	  }
	else
	  {
	    return(-((TheNumeric)(-2.0*x + 1.0) / 2));
	  }
}

void shortvec(TheNumeric **F, int max, veclist *V)
/*****	calculates the vectors of norm <= max 
      with respect to the Gram-matrix F, 
      writes the vectors on V->v and the 
      length on V->v[i][dim1],
      the algorithm uses the model of the 
      LLL-algorithm	*****/
{
	int     dim1, i, j, m, l, rank, con;
	double	**mu, *B;
	TheNumeric **gram, **T, *vec;

	dim1 = V->dim1;
	if ((B = (double*)malloc(dim1 * sizeof(double))) == 0)
		exit (1);
	if ((mu = (double**)malloc(dim1 * sizeof(double*))) == 0)
		exit (1);
/* F is copied on gram and remains unchanged */
	if ((gram = (TheNumeric**)malloc(dim1 * sizeof(TheNumeric*))) == 0)
		exit (1);
	if ((T = (TheNumeric**)malloc(dim1 * sizeof(TheNumeric*))) == 0)
		exit (1);
	if ((vec = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
		exit (1);
	for (i = 0; i < dim1; ++i)
	{
		vec[i] = 0;
		if ((mu[i] = (double*)malloc(dim1 * sizeof(double))) == 0)
			exit (1);
		if ((gram[i] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
			exit (1);
		if ((T[i] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
			exit (1);
		for (j = 0; j < dim1; ++j)
		{
			T[i][j] = 0;
			mu[i][j] = 0.0;
			gram[i][j] = F[i][j];
		}
		T[i][i] = 1;
		mu[i][i] = 1.0;
	}
/* no basis-vector should have norm 0 */
	for (i = 0; i < dim1  &&  gram[i][i] > 0; ++i);
	if (i != dim1)
	{
		fprintf(stderr, "Error: matrix is not positive definite\n");
		exit (3);
	}
	rank = dim1;
	initialize(0, mu, B, gram);
	if (dim1 > 1)
		initialize(1, mu, B, gram);
/* do the LLL-reduction */
	m = 1;
	l = 0;
	while (m < rank  &&  rank == dim1)
		m = red(&m, &l, gram, T, mu, B, dim1, &rank);
	if (rank < dim1)
	{
		fprintf(stderr, "Error: matrix is not positive definite\n");
		exit (3);
	}
	if ((V->v = (TheNumeric**)malloc(sizeof(TheNumeric*))) == 0)
		exit (1);
	if ((V->v[0] = (TheNumeric*)malloc(V->len * sizeof(TheNumeric))) == 0)
		exit (1);
	V->n = 0;
	con = 0;
/* recursively find the vectors up to length max */
	shrt(dim1-1, 0.0, max, V, vec, mu, T, B, &con);
	free(B);
	free(vec);
	for (i = 0; i < dim1; ++i)
	{
		free(mu[i]);
		free(T[i]);
		free(gram[i]);
	}
	free(mu);
	free(T);
	free(gram);
}

void shrt(int k, double norm, int max, veclist *V, TheNumeric *vec, double **mu, TheNumeric **T, double *B, int *con)
/*****	recursion for finding the vectors up 
      to length max, where the components
      vec[k+1],...,vec[dim1-1] are already
      chosen and contribute norm to the 
      length of the vector	*****/
{
  double	x, Bk;
  int	i, j, dim1;
  TheNumeric *Vvn, Vvni;
  dim1 = V->dim1;
  if (k == -1)
    {
      for (i = 0; i < dim1  &&  vec[i] == 0; ++i);
      if (i == dim1)
	/* the 0-vector indicates the end of the recursion, since the short vectors
	   are only enumerated up to sign */
	*con = 1;
      else
	/* a short vector is found */
	{
	  ++V->n;
	  if ((V->v = (TheNumeric**)realloc(V->v, (V->n+1) * sizeof(TheNumeric*))) == 0)
	    exit (1);
	  if ((V->v[V->n] = (TheNumeric*)malloc(V->len * sizeof(TheNumeric))) == 0)
	    exit (1);
	  Vvn = V->v[V->n];
	  for (i = 0; i < dim1; ++i)
	    {
	      Vvni = 0;
	      for (j = 0; j < dim1; ++j)
		Vvni += vec[j] * T[j][i];
	      Vvn[i] = Vvni;
	    }
	  Vvn[dim1] = MyRound(norm);
	}
    }
  else
    {
      x = 0.0;
      for (j = k+1; j < dim1; ++j)
			x += vec[j] * mu[j][k];
      i = -MyRound(x);	
      Bk = B[k];
      if (Bk*(x+i)*(x+i) + norm < max + EPS)
	{
	  while ((Bk*(x+i)*(x+i) + norm <= max + EPS) || 
		 (x+i <= 0))
	    ++i;
	  --i;
	  while ((Bk*(x+i)*(x+i) + norm < max + EPS) && *con == 0)
	    {
	      vec[k] = i;
	      /* go deeper into the recursion */
	      shrt(k-1, Bk*(x+i)*(x+i)+norm, max, V, vec, mu, T, B, con);
	      --i;
	    }
	}
    }
//	free(Vvn);
}

void change(TheNumeric **v, int nv, TheNumeric **T, int n)
/*****	changes v according to the 
      transformation matrix T,
      i.e. v := T*v	*****/
{
  int	i, j, k, fac;
  TheNumeric **w, *wi, *Ti, *vi, *vj;
  if ((w = (TheNumeric**)malloc(nv * sizeof(TheNumeric*))) == 0)
    exit (1);
  for (i = 0; i < nv; ++i)
    {
      if ((w[i] = (TheNumeric*)malloc(n * sizeof(TheNumeric))) == 0)
	exit (1);
      wi = w[i];
      for (j = 0; j < n; ++j)
	wi[j] = 0;
    }
  for (i = 0; i < nv; ++i)
    {
      wi = w[i];
      Ti = T[i];
      for (j = 0; j < nv; ++j)
	{
	  if ((fac = Ti[j]) != 0)
	    {
	      vj = v[j];
	      for (k = 0; k < n; ++k)
		wi[k] += fac * vj[k];
	    }
	}
    }
  for (i = 0; i < nv; ++i)
    {
      vi = v[i];
      wi = w[i];
      for (j = 0; j < n; ++j)
	vi[j] = wi[j];
      free(wi);
    }
  free(w);
}

void vecmatmul(TheNumeric *x, TheNumeric **A, int n, TheNumeric *y)
{
  int	i, j;
  TheNumeric *Ai, xi;
  for (j = 0; j < n; ++j)
    y[j] = 0;
  for (i = 0; i < n; ++i)
    {
      if ((xi = x[i]) != 0)
	{
	  Ai = A[i];
	  for (j = 0; j < n; ++j)
	    y[j] += xi * Ai[j];
	}
    }
}

void matmul(TheNumeric **A, TheNumeric **B, int n, TheNumeric **C)
/*****	multiplies nxn-matrices A and B and puts result
      on C	*****/
{
  int	i, j, k;
  TheNumeric *Ai, *Bk, *Ci, Aik;
  for (i = 0; i < n; ++i)
    {
      Ai = A[i];
      Ci = C[i];
      for (j = 0; j < n; ++j)
	Ci[j] = 0;
      for (k = 0; k < n; ++k)
	{
	  if ((Aik = Ai[k]) != 0)
	    {
	      Bk = B[k];
	      for (j = 0; j < n; ++j)
		Ci[j] += Aik * Bk[j];
	    }
	}
    }
}

TheNumeric scp(TheNumeric *x, TheNumeric **F, TheNumeric *y, int n)
/*****	returns scalar product of 1xn-vectors x and y
      with respect to Gram-matrix F	*****/
{
  int	i, j;
  TheNumeric sum, *Fi, xi;
  sum = 0;
  for (i = 0; i < n; ++i)
    {
      if ((xi = x[i]) != 0)
	{
	  Fi = F[i];
	  for (j = 0; j < n; ++j)
	    sum += xi * Fi[j] * y[j];
	}
    }
  return(sum);
}

TheNumeric sscp(TheNumeric *x, TheNumeric *y, int n)
/*****	returns standard scalar product of 
      1xn-vectors x and y, heavily used	*****/
{
	int	i;
	TheNumeric sum;
	sum = 0;
	for (i = 0; i < n; ++i)
		sum += *(x++) * *(y++);
	/*	sum += x[i] * y[i];	*/
	return(sum);
}

void psolve(TheNumeric **X, TheNumeric **A, TheNumeric **B, int n, TheNumeric p)
/*****	computes X = B*A^-1 modulo the prime p
      for nxn-matrices A and B, where A is
      invertible,
      A and B are modified !!!	*****/
{
	int	i, j, k;
	TheNumeric *Xi, *Bi, ainv, sum, tmp;
/* convert A to lower triangular matrix and change B accordingly */
	for (i = 0; i < n-1; ++i)
	{
		for (j = i; A[i][j] % p == 0; ++j);
		if (j == n)
		{
		  fprintf(stderr, "Error: matrix is singular modulo ");
		  single2Print(stderr, p);
		  fprintf(stderr, "\n");
		  exit (2);
		}
		if (j != i)
/* interchange columns i and j such that A[i][i] is != 0 */
		{
			for (k = i; k < n; ++k)
			{
				tmp = A[k][i];
				A[k][i] = A[k][j];
				A[k][j] = tmp;
			}
			for (k = 0; k < n; ++k)
			{
				tmp = B[k][i];
				B[k][i] = B[k][j];
				B[k][j] = tmp;
			}
		}
		pgauss(i, A, B, n, p);
	}
/* compute X recursively */
	for (i = 0; i < n; ++i)
	{
		Xi = X[i];
		Bi = B[i];
		for (j = n-1; j >= 0; --j)
		{
			sum = 0;
			for (k = n-1; k > j; --k)
				sum = (sum + Xi[k] * A[k][j]) % p;
/* ainv is the inverse of A[j][j] modulo p */
			for (ainv = 1; abs(A[j][j]*ainv-1) % p != 0; ++ainv);
			Xi[j] = (Bi[j] - sum) * ainv % p;
/* make sure that -p/2 < X[i][j] <= p/2 */
			if (2*Xi[j] > p)
				Xi[j] -= p;
			else if (2*Xi[j] <= -p)
				Xi[j] += p;
		}
	}
}

void pgauss(int r, TheNumeric **A, TheNumeric **B, int n, TheNumeric p)
/*****	clears row nr. r of A assuming that the 
      first r-1 rows of A are already cleared 
      and that A[r][r] != 0 modulo p, 
      A and B	are changed !!!	*****/
{
	int	i, j;
	TheNumeric *Ar, f, ainv;

	Ar = A[r];
/* ainv is the inverse of A[r][r] modulo p */
	for (ainv = 1; abs(Ar[r]*ainv-1) % p != 0; ++ainv);
	for (j = r+1; j < n; ++j)
	{
		if (Ar[j] % p != 0)
		{
			f = Ar[j] * ainv % p;
			for (i = r+1; i < n; ++i)
				A[i][j] = (A[i][j] - f * A[i][r]) % p;
			for (i = 0; i < n; ++i)
				B[i][j] = (B[i][j] - f * B[i][r]) % p;
			Ar[j] = 0;
		}
	}
}

int isprime(TheNumeric n)
/*****	checks, whether n is a prime	*****/
{
	int	i;

	for (i = 2; i <= n/i; ++i)
	{
		if (n % i == 0)
			return 0;
	}
	return 1;
}

int operate(int nr, TheNumeric **A, veclist V)
/*****	V.v is a sorted list of length V.n of vectors
      of dim1ension V.dim1, the number of V.v[nr]*A in
      the list is returned, where a negative number 
      indicates the negative of a vector	*****/
{
	int	i, im;
	TheNumeric *w;
	if ((w = (TheNumeric*)malloc(V.dim1 * sizeof(TheNumeric))) == 0)
		exit (1);
	vecmatmul(V.v[abs(nr)], A, V.dim1, w);
	if (nr < 0)
	{
		for (i = 0; i < V.dim1; ++i)
			w[i] *= -1;
	}
	im = numberof(w, V);
	if (abs(im) > V.n)
/* the vector is not in the list */
	{
		fprintf(stderr, "Error: image of vector %d not found\n", nr);
		exit (2);
	}
	free(w);
	return(im);
}

int orbit(int	*pt, int npt, TheNumeric ***G, int nG, veclist V, int **orb)
/*****	computes the orbit of npt points in pt 
      under the nG matrices in G and puts the
      orbit in orb, allocates the memory for 
      orb, returns the length of the orbit, 
      the points are the indices in the
      list V.v	*****/
{
	int	i, norb, cnd, im, *flag;

	if ((*orb = (int*)malloc(npt * sizeof(int))) == 0)
		exit (1);
/* if flag[i + V.n] = 1, then the point i is already in the orbit */
	if ((flag = (int*)malloc((2*V.n + 1) * sizeof(int))) == 0)
		exit (1);
	for (i = 0; i <= 2*V.n; ++i)
		flag[i] = 0;
	for (i = 0; i < npt; ++i)
	{
		(*orb)[i] = pt[i];
		flag[pt[i]+V.n] = 1;
	}
	norb = npt;
	cnd = 0;
	while (cnd < norb)
	{
		for (i = 0; i < nG; ++i)
		{
			im = operate((*orb)[cnd], G[i], V);
			if (flag[im+V.n] == 0)
/* the image is a new point in the orbit */
			{
				++norb;
				if ((*orb = (int*)realloc(*orb, norb * sizeof(int))) == 0)
					exit (1);
				(*orb)[norb-1] = im;
				flag[im+V.n] = 1;
			}
		}
		++cnd;
	}
	free(flag);
	return(norb);
}

int orbitlen(int pt, int orblen, TheNumeric ***G, int nG, veclist V)
/*****	checks, whether the orbit of pt under 
      the nG matrices in G has at least length
      orblen	*****/
{
	int	i, len, cnd, im, *orb, *flag;

	if ((orb = (int*)malloc(orblen * sizeof(int))) == 0)
		exit(1);
/* if flag[i + V.n] = 1, then the point i is already in the orbit */
	if ((flag = (int*)malloc((2*V.n + 1) * sizeof(int))) == 0)
		exit(1);
	for (i = 0; i <= 2*V.n; ++i)
		flag[i] = 0;
	orb[0] = pt;
	flag[pt+V.n] = 1;
	for (i = 1; i < orblen; ++i)
		orb[i] = 0;
	len = 1;
	cnd = 0;
	while (cnd < len  &&  len < orblen)
	{
		for (i = 0; i < nG  &&  len < orblen; ++i)
		{
			im = operate(orb[cnd], G[i], V);
			if (flag[im+V.n] == 0)
/* the image is a new point in the orbit */
			{
				orb[len] = im;
				++len;
				flag[im+V.n] = 1;
			}
		}
		++cnd;
	}
	free(orb);
	free(flag);
	return(len);
}

int delete_new(int *orb1, int l1, int *orb2, int l2)
/*****	deletes the elements in orb2 from orb1, 
      an entry 0 marks the end of the list, 
      returns the length of orb1	*****/
{
	int	i, j, len, o2i;

/* the true length of orb1 is len, hence orb1[len-1] != 0 */
	for (i = 0; i < l1  &&  orb1[i] != 0; ++i);
	len = i;
	for (i = 0; i < l2  &&  orb2[i] != 0; ++i)
	{
		o2i = orb2[i];
		for (j = 0; j < len  &&  orb1[j] != o2i; ++j);
/* orb1[j] = orb2[i], hence delete orb1[j] from orb1 */
		if (j < len)
		{
			orb1[j] = orb1[len-1];
			orb1[len-1] = 0;
			--len;
		}	
	}
	return(len);
}

void stab(int I, group *G, fpstruct fp, veclist V)
/*****	computes the orbit of fp.e[I] under the 
      generators in G->g[I]...G->g[n-1] and elements 
      stabilizing fp.e[I],
      has some heuristic break conditions,
      the generators in G->g[i] stabilize 
      fp.e[0]...fp.e[i-1] but not fp.e[i], 
      G->ng[i] is the number of generators in G->g[i],
      the first G->nsg[i] of which are elements which
      are obtained as stabilizer elements in 
      <G->g[0],...,G->g[i-1]>, G->ord[i] is the orbit
      length of fp.e[i] under 
      <G->g[i],...,G->g[n-1]>	*****/
{
	int	*orb, len, cnd, tmplen;
	int	*flag, **w;
	int	i, j, k, l, dim1, im, nH, nHj, fail;
	int	Maxfail, Rest;
	TheNumeric ***H, ***Hj, **S, ***Ggj, **tmp;
/* some heuristic break conditions for the computation of stabilizer elements:
   it would be too expensive to calculate all the stabilizer generators, which
   are obtained from the orbit, since this is highly redundant, 
   on the other hand every new generator which enlarges the group is much 
   cheaper than one obtained from the backtrack,
   after Maxfail subsequent stabilizer elements, that do not enlarge the group,
   Rest more elements are calculated even if they leave the group unchanged,
   since it turned out that this is often useful in the following steps,
   increasing the parameters will possibly decrease the number of generators
   for the group, but will increase the running time,
   there is no magic behind this heuristic, tuning might be appropriate */
	dim1 = V.dim1;
	Rest = 0;
	for (i = I; i < dim1; ++i)
	{
		if (fp.diag[i] > 1  &&  G->ord[i] < fp.diag[i])
			++Rest;
	}
	Maxfail = Rest;
	for (i = 0; i < dim1; ++i)
	{
		if (fp.diag[i] > 1)
			++Maxfail;
	}
	nH = 0;
	for (i = I; i < dim1; ++i)
		nH += G->ng[i];
/* H are the generators of the group in which the stabilizer is computed */
	if ((H = (TheNumeric***)malloc(nH * sizeof(TheNumeric**))) == 0)
		exit (1);
	if ((Hj = (TheNumeric***)malloc((nH+1) * sizeof(TheNumeric**))) == 0)
		exit (1);
	k = 0;
	for (i = I; i < dim1; ++i)
	{
		for (j = 0; j < G->ng[i]; ++j)
		{
			H[k] = G->g[i][j];
			++k;
		}
	}
/* in w[V.n+i] an element is stored that maps fp.e[I] on v[i] */
	if ((w = (int**)malloc((2*V.n+1) * sizeof(int*))) == 0)
		exit (1);
/* orb contains the orbit of fp.e[I] */
	if ((orb = (int*)malloc(2*V.n * sizeof(int))) == 0)
		exit (1);
	for (i = 0; i < 2*V.n; ++i)
		orb[i] = 0;
/* if flag[i + V.n] = 1, then the point i is already in the orbit */
	if ((flag = (int*)malloc((2*V.n+1) * sizeof(int))) == 0)
		exit (1);
	for (i = 0; i <= 2*V.n; ++i)
		flag[i] = 0;
/* S is a matrix to hold a stabilizer element temporarily */
	if ((S = (TheNumeric**)malloc(dim1 * sizeof(TheNumeric*))) == 0)
		exit (1);
	for (i = 0; i < dim1; ++i)
	{
		if ((S[i] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
			exit (1);
	}
	orb[0] = fp.e[I];
	flag[orb[0]+V.n] = 1;
	if ((w[orb[0]+V.n] = (int*)malloc(dim1 * sizeof(int))) == 0)
		exit (1);
	for (i = 0; i < dim1; ++i)
		w[orb[0]+V.n][i] = fp.e[i];
	cnd = 0;
	len = 1;
/* fail is the number of successive failures */
	fail = 0;
	while (len > cnd  &&  fail < Maxfail+Rest)
	{
		for (i = 0; i < nH  &&  fail < Maxfail+Rest; ++i)
		{
			if (fail >= Maxfail)
/* there have already been Maxfail successive failures, now a random generator
   is applied to a random point of the orbit to get Rest more stabilizer 
   elements */
			{
				cnd = rand() % len;
				if (cnd < 0)
					cnd += len;
				i = rand() % nH;
				if (i < 0)
					i += nH;
			}
			im = operate(orb[cnd], H[i], V);
			if (flag[im+V.n] == 0)
/* a new element is found, appended to the orbit and an element mapping
   fp.e[I] to im is stored in w[im+V.n] */
			{
				orb[len] = im;
				flag[im+V.n] = 1;
				if ((w[im+V.n] = (int*)malloc(dim1 * sizeof(int))) == 0)
					exit (1);
				for (j = 0; j < dim1; ++j)
					w[im+V.n][j] = operate(w[orb[cnd]+V.n][j], H[i], V);
				++len;
			}
			else
/* the image was already in the orbit */
			{
/* j is the first index where the images of the old and the new element
   mapping e[I] on im differ */
				for (j = I; j < dim1  &&  operate(w[orb[cnd]+V.n][j], H[i], V) == w[im+V.n][j]; ++j);
				if (j < dim1  &&  
				    (G->ord[j] < fp.diag[j] || fail >= Maxfail))
				{
/* new stabilizer element S = w[orb[cnd]+V.n] * H[i] * (w[im+V.n])^-1 */
					stabil(S, w[orb[cnd]+V.n], w[im+V.n], fp.per, H[i], V);
					Hj[0] = S;
					nHj = 1;
					for (k = j; k < dim1; ++k)
					{
						for (l = 0; l < G->ng[k]; ++l)
						{
							Hj[nHj] = G->g[k][l];
							++nHj;
						}
					}
					tmplen = orbitlen(fp.e[j], fp.diag[j], Hj, nHj, V);
					if (tmplen > G->ord[j]  ||  fail >= Maxfail)
/* the new stabilizer element S either enlarges the orbit of e[j]
   or it is one of the additional elements after MAXFAIL failures */
					{
						G->ord[j] = tmplen;
						++G->ng[j];
						++G->nsg[j];
						/* allocate memory for the new generator */
						if ((G->g[j] = (TheNumeric***)realloc(G->g[j], G->ng[j] * sizeof(TheNumeric**))) == 0)
							exit (1);
						if ((G->g[j][G->ng[j]-1] = (TheNumeric**)malloc(dim1 * sizeof(TheNumeric*))) == 0)
							exit (1);
						for (k = 0; k < dim1; ++k)
						{
							if ((G->g[j][G->ng[j]-1][k] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
								exit (1);
						}
						Ggj = G->g[j];
/* the new generator is inserted as stabilizer element nr. nsg[j]-1 */
						for (k = G->ng[j]-1; k >= G->nsg[j]; --k)
						{
							tmp = Ggj[k];
							Ggj[k] = Ggj[k-1];
							Ggj[k-1] = tmp;
						}
						for (k = 0; k < dim1; ++k)
						{
							for (l = 0; l < dim1; ++l)
								Ggj[G->nsg[j]-1][k][l] = S[k][l];
						}
						++nH;
						if ((H = (TheNumeric***)realloc(H, nH * sizeof(TheNumeric**))) == 0)
							exit (1);
						if ((Hj = (TheNumeric***)realloc(Hj, (nH+1) * sizeof(TheNumeric**))) == 0)
							exit (1);
/* the new generator is appended to H */
						H[nH-1] = Ggj[G->nsg[j]-1];
/* the number of failures is reset to 0 */ 
						if (fail < Maxfail)
							fail = 0;
						else
							++fail;
					}
					else
/* the new stabilizer element S does not enlarge the orbit of e[j] */
						++fail;
				}
				else if ((j < dim1  &&  fail < Maxfail)  ||  
					 (j == dim1  &&  fail >= Maxfail))
					++fail;
/* if S is the identity and fail < Maxfail, nothing is done */
			}
		}
		if (fail < Maxfail)
			++cnd;
	}
	for (i = 0; i <= 2*V.n; ++i)
	{
		if (flag[i] == 1)
			free(w[i]);
	}
	free(w);
	for (i = 0; i < dim1; ++i)
		free(S[i]);
	free(S);
	free(orb);
	free(flag);
	free(H);
	free(Hj);
}

void matgen(int *x, TheNumeric **X, int dim1, int *per, TheNumeric **v)
/*****	generates the matrix X which has as row
      per[i] the vector nr. x[i] from the 
      list v	*****/
{
  int	i, j, xi;
  TheNumeric *Xperi;
  for (i = 0; i < dim1; ++i)
    {
      Xperi = X[per[i]];
      if ((xi = x[i]) > 0)
	{
	  for (j = 0; j < dim1; ++j)
	    Xperi[j] = v[xi][j];
	}
      else
	{
	  for (j = 0; j < dim1; ++j)
	    Xperi[j] = -v[-xi][j];
	}
    }
}

void stabil(TheNumeric **S, int *x1, int *x2, int *per, TheNumeric **G, veclist V)
/*****	x1 corresponds to an element X1
      mapping some vector e on p1,
      x2 to an element X2 mapping e on
      p2 and G is a generator mapping
      p1 on p2, then S = X1*G*X2^-1
      stabilizes e	*****/
{
  int	i, dim1, *x;
  TheNumeric **XG, **X2;
  dim1 = V.dim1;
  if ((XG = (TheNumeric**)malloc(dim1 * sizeof(TheNumeric*))) == 0)
    exit (1);
  if ((X2 = (TheNumeric**)malloc(dim1 * sizeof(TheNumeric*))) == 0)
    exit (1);
  for (i = 0; i < dim1; ++i)
    {
      if ((XG[i] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
	exit (1);
      if ((X2[i] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
	exit (1);
    }
  if ((x = (int*)malloc(dim1 * sizeof(int))) == 0)
    exit (1);
  for (i = 0; i < dim1; ++i)
    x[i] = operate(x1[i], G, V);
  /* XG is the composite mapping of the matrix corresponding to x1 and G */
  matgen(x, XG, dim1, per, V.v);
  matgen(x2, X2, dim1, per, V.v);
  /* S = XG * X2^-1 */
  psolve(S, X2, XG, dim1, V.prime);
  free(x);
  for (i = 0; i < dim1; ++i)
    {
      free(XG[i]);
      free(X2[i]);
    }
  free(XG);
  free(X2);
}

void checkvecs(veclist *V, invar F, veclist norm)
/*****	removes those vectors from V.v, which 
      have other norm combinations than the 
      base-vectors	*****/
{
	int	i, j, k, dim1;
	TheNumeric *Vvi, *normvec;
	dim1 = F.dim1;
	if ((normvec = (TheNumeric*)malloc(norm.dim1 * sizeof(TheNumeric))) == 0)
		exit (EXIT_FAILURE);
	j = 0;
	for (i = 1; i <= V->n; ++i)
	{
		Vvi = V->v[i];
		for (k = 0; k < 1; ++k)
		{
			Vvi[dim1+k] = scp(Vvi, F.A[k], Vvi, dim1);
			normvec[k] = Vvi[dim1+k];
		}
		if (abs(numberof(normvec, norm)) > dim1)
/* the vector V->V[i] has a different norm combination from those in the list
   norm.v and is therefore deleted */
		{
			free(Vvi);
			++j;
		}
		else
			V->v[i-j] = Vvi;
	}
	V->n -= j;
	free(normvec);
}

int checkgen(TheNumeric **g, invar F)
/*****	checks, whether g fixes the invariant forms	*****/
{
  int	i, j, k, fix;
  TheNumeric **FAk;
  fix = 1;
  for (k = 0; k < 1  &&  fix == 1; ++k)
    {
      FAk = F.A[k];
      for (i = 0; i < F.dim1  &&  fix == 1; ++i)
	{
	  for (j = 0; j <= i  &&  fix == 1; ++j)
	    {
	      if (scp(g[i], FAk, g[j], F.dim1) != FAk[i][j])
		fix = 0;
	    }
	}
    }
  return(fix);
}

void fingerprint(fpstruct *fp, invar F, veclist	V)
/*****	a permutation per := fp.per for the
      order of the basis-vectors is chosen
      such that in every step the number of
      possible continuations is minimal,
      for j from per[i] to per[dim1-1] the
      value f[i][j] in the fingerprint f is
      the number of vectors, which have the
      same scalar product with the 
      basis-vectors per[0]...per[i-1] as the 
      basis-vector j and the same length as 
      this vector with respect to all 
      invariant forms	*****/
{
	int	i, j, k, dim1, min, tmp, **f;
	TheNumeric *Vvj, ***Ftr;

	dim1 = F.dim1;
	if ((Ftr = (TheNumeric***)malloc(sizeof(TheNumeric**))) == 0)
	  exit(EXIT_FAILURE);
	for (i = 0; i < 1; ++i)
	{
		if ((Ftr[i] = (TheNumeric**)malloc(dim1 * sizeof(TheNumeric*))) == 0)
			exit (EXIT_FAILURE);
		for (j = 0; j < dim1; ++j)
		{
			if ((Ftr[i][j] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
				exit (EXIT_FAILURE);
			for (k = 0; k < dim1; ++k)
				Ftr[i][j][k] = F.A[i][k][j];
		}
	}
	if ((f = (int**)malloc(dim1 * sizeof(int*))) == 0)
		exit (EXIT_FAILURE);
	for (i = 0; i < dim1; ++i)
	{
		if ((f[i] = (int*)malloc(dim1 * sizeof(int))) == 0)
			exit (EXIT_FAILURE);
		for (j = 0; j < dim1; ++j)
			f[i][j] = 0;
	}
	for (i = 0; i < dim1; ++i)
		fp->per[i] = i;
/* the first row of the fingerprint has as entry nr. i the number of
   vectors, which have the same length as the i-th basis-vector with
   respect to every invariant form */
	for (j = 1; j <= V.n; ++j)
	{
		Vvj = V.v[j];
		for (i = 0; i < dim1; ++i)
		{
			for (k = 0; k < 1  &&  Vvj[dim1+k] == F.A[k][i][i]; ++k);
			if (k == 1)
				f[0][i] += 2;
		}
	}
	for (i = 0; i < dim1-1; ++i)
	{
/* a minimal entry != 0 in the i-th row is chosen */
		min = i;
		for (j = i+1; j < dim1; ++j)
		{
			if (f[i][fp->per[j]] < f[i][fp->per[min]])
				min = j;
		}
		tmp = fp->per[i];
		fp->per[i] = fp->per[min];
		fp->per[min] = tmp;
/* the column below the minimal entry is set to 0 */
		for (j = i+1; j < dim1; ++j)
			f[j][fp->per[i]] = 0;
/* compute the row i+1 of the fingerprint */
		for (j = i+1; j < dim1; ++j)
		  f[i+1][fp->per[j]] = possible(F, Ftr, V, fp->per, i, fp->per[j]);
	}
/* only the diagonal of f will be needed later */
	for (i = 0; i < dim1; ++i)
	{
		fp->diag[i] = f[i][fp->per[i]];
		free(f[i]);
	}
	free(f);
	for (i = 0; i < 1; ++i)
	{
		for (j = 0; j < dim1; ++j)
			free(Ftr[i][j]);
		free(Ftr[i]);
	}
	free(Ftr);
}

int possible(invar F, TheNumeric ***Ftr, veclist V, int *per, int I, int J)
/*****	returns the number of vectors, which
      have the same scalar products with the
      basis-vectors nr. per[0]...per[I] as the
      basis-vector nr. J and the same length
      as this vector with respect to all
      invariant forms	*****/
{
	int	i, j, k, dim1, count;
	TheNumeric *Vvj;
	dim1 = F.dim1;
	count = 0;
	for (j = 1; j <= V.n; ++j)
	{
		Vvj = V.v[j];
		i = I+1;
/* check the length of the vector */
		for (k = 0; k < 1  &&  i > I  &&  Vvj[dim1+k] == F.A[k][J][J]; ++k)
/* check the scalar products with the basis-vectors */
		/*	for (i = 0; i <= I  &&  F.v[k][j][per[i]] == F.A[k][per[i]][J]; ++i); */
			for (i = 0; i <= I  &&  sscp(Vvj, Ftr[k][per[i]], dim1) == F.A[k][J][per[i]]; ++i);
		if (k == 1  &&  i > I)
			++count;
/* the same for the negative vector */
		i = I+1;
		for (k = 0; k < 1  &&  i > I  &&  Vvj[dim1+k] == F.A[k][J][J]; ++k)
		/*	for (i = 0; i <= I  &&  F.v[k][j][per[i]] == -F.A[k][per[i]][J]; ++i); */
			for (i = 0; i <= I  &&  sscp(Vvj, Ftr[k][per[i]], dim1) == -F.A[k][J][per[i]]; ++i);
		if (k == 1  &&  i > I)
			++count;
	}
	return(count);
}

void scpvector(TheNumeric *scpvec, TheNumeric *w, int *b, int I, int dep, invar	F)
/*****	calculates the scalar products
      of the vector w with the base
      vectors v[b[I]] down to
      v[b[I-dep+1]] with respect to
      all invariant forms and puts
      them on scpvec	*****/
{
  int	i, j, dim1, bi;
  
  dim1 = F.dim1;
  for (i = I; i >= 0  &&  i > I-dep; --i)
    {
      if ((bi = b[i]) > 0)
	{
	  for (j = 0; j < 1; ++j)
	    scpvec[j*dep + I-i] = sscp(w, F.v[j][bi], dim1);
	}
      else
	{
	  for (j = 0; j < 1; ++j)
	    scpvec[j*dep + I-i] = -sscp(w, F.v[j][-bi], dim1);
	}
    }
}

void scpvecs(veclist *list, TheNumeric ***vec, int I, int *b, int dep, veclist V, invar F)
/*****	computes the list of scalar product 
      combinations of the vectors in V.v with
      the basis-vectors in b	*****/
{
	int	i, j, dim1, len, nr, sign;
	TheNumeric *Vvj, *listvn, **listv, *tmp;
	TheNumeric *vecnr, *vecn, *scpvec;

	dim1 = F.dim1;
	len = dep;
/* scpvec is the vector for the scalar product combination */
	if ((scpvec = (TheNumeric*)malloc(len * sizeof(TheNumeric))) == 0)
		exit (EXIT_FAILURE);
/* the first vector in the list is the 0-vector and is not counted */
	if ((list->v = (TheNumeric**)malloc(1 * sizeof(TheNumeric*))) == 0)
		exit (EXIT_FAILURE);
	list->dim1 = len;
	list->len = len;
	if ((list->v[0] = (TheNumeric*)malloc(len * sizeof(TheNumeric))) == 0)
		exit (EXIT_FAILURE);
	for (i = 0; i < len; ++i)
		list->v[0][i] = 0;
	list->n = 0;
	if ((*vec = (TheNumeric**)malloc(1 * sizeof(TheNumeric*))) == 0)
		exit (EXIT_FAILURE);
	if (((*vec)[0] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
		exit (EXIT_FAILURE);
	for (i = 0; i < dim1; ++i)
		(*vec)[0][i] = 0;
	for (j = 1; j <= V.n; ++j)
	{
		Vvj = V.v[j];
		for (i = 0; i < len; ++i)
			scpvec[i] = 0;
		scpvector(scpvec, Vvj, b, I, dep, F);
		for (i = 0; i < len  &&  scpvec[i] == 0; ++i);
		if (i == len)
/* if scpvec is the 0-vector, nr is set to 0, since numberof never returns 0 */
			nr = 0;
		else
		{
			nr = numberof(scpvec, *list);
			sign = nr > 0 ? 1 : -1;
			nr = abs(nr);
		}
/* scpvec is already in list */
		if (nr <= list->n  &&  nr > 0)
		{
			vecnr = (*vec)[nr];
			for (i = 0; i < dim1; ++i)
				vecnr[i] += sign * Vvj[i];
		}
/* scpvec is a new scalar product combination */
		else if (nr >= list->n+1)
		{
			++list->n;
			if ((list->v = (TheNumeric**)realloc(list->v, (list->n+1) * sizeof(TheNumeric*))) == 0)
				exit (EXIT_FAILURE);
			if ((list->v[list->n] = (TheNumeric*)malloc(len * sizeof(TheNumeric))) == 0)
				exit (EXIT_FAILURE);
/* numberof changes the sign of scpvec if the first nonzero entry is < 0,
   hence this is correct */
			listvn = list->v[list->n];
			for (i = 0; i < len; ++i)
				listvn[i] = scpvec[i];
			if ((*vec = (TheNumeric**)realloc(*vec, (list->n+1) * sizeof(TheNumeric*))) == 0)
				exit (EXIT_FAILURE);
			if (((*vec)[list->n] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
				exit (EXIT_FAILURE);
			vecn = (*vec)[list->n];
			for (i = 0; i < dim1; ++i)
				vecn[i] = sign * Vvj[i];
/* shuffle the new vector to the correct position, this should be quick enough,
   since the length of the list should not exceed some hundreds */
			listv = list->v;
			for (i = list->n; i > nr+1-list->n; --i)
			{
				tmp = listv[i];
				listv[i] = listv[i-1];
				listv[i-1] = tmp;
				tmp = (*vec)[i];
				(*vec)[i] = (*vec)[i-1];
				(*vec)[i-1] = tmp;
			}
		}
	}
	free(scpvec);
}

void base(scpcomb *com, TheNumeric ***b, TheNumeric **v, TheNumeric **F, int dim1)
/*****	computes a basis b for the lattice 
      generated by the vectors in v, puts a 
      transformation matrix on T, 
      i.e. b = T*v, uses LLL-reduction with 	
      respect to F	*****/
{
	int	i, j, k, nv, rank, *perm, tmp;
	TheNumeric *Fvi, **Fv, *vi, **f, *fj, **tr, *comtr, *norm;
	TheNumeric TheMaxNorm, max;
	TheMaxNorm=GetMAXNORM();
	nv = com->list.n + 1;
	max = 1;
/* get the maximal entry in the vector sums */
	for (i = 0; i < nv; ++i)
	{
		for (j = 0; j < dim1; ++j)
		{
			if (abs(v[i][j]) > max)
				max = abs(v[i][j]);
		}
	}
/* Fv is the list of products of F with the vectors in v, scalar products
   with respect to F can be calculated as standard scalar products of
   vectors in v with vectors in Fv */
	if ((Fv = (TheNumeric**)malloc(nv * sizeof(TheNumeric*))) == 0)
		exit (EXIT_FAILURE);
/* the product of the maximal entry in the vector sums with the maximal entry
   in the product of the Gram-matrices with the vector sums should not exceed
   MAXNORM to avoid overflow */
	max = TheMaxNorm / max;
	for (i = 0; i < nv; ++i)
	{
		if ((Fv[i] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
			exit (EXIT_FAILURE);
		Fvi = Fv[i];
		for (j = 0; j < dim1; ++j)
		{
			Fvi[j] = sscp(F[j], v[i], dim1);
			if (abs(Fvi[j]) > max)
/* an entry in F.v[i] is too large */
			{
			  fprintf(stderr, "Error: Found entry ");
			  single2Print(stderr, Fvi[j]);
			  fprintf(stderr," in vector sum.\nTo avoid overflow, the entries should not exceed ");
			  single2Print(stderr, max);
			  fprintf(stderr, ".\nTry without option -D or use -D with higher value.\n");
			  exit (3);
			}
		}
	}
/* b is the basis for the lattice */
	if ((*b = (TheNumeric**)malloc(1 * sizeof(TheNumeric*))) == 0)
		exit (EXIT_FAILURE);
	if (((*b)[0] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
		exit (EXIT_FAILURE);
/* com->trans is the transformation matrix */
	if ((com->trans = (TheNumeric**)malloc(1 * sizeof(TheNumeric*))) == 0)
		exit (EXIT_FAILURE);
	if ((com->trans[0] = (TheNumeric*)malloc(nv * sizeof(TheNumeric))) == 0)
		exit (EXIT_FAILURE);
/* f is the Gram-matrix for the basis b with respect to the form F */
	if ((f = (TheNumeric**)malloc(1 * sizeof(TheNumeric*))) == 0)
		exit (EXIT_FAILURE);
	if ((f[0] = (TheNumeric*)malloc(1 * sizeof(TheNumeric))) == 0)
		exit (EXIT_FAILURE);
/* tr is the transformation matrix for the LLL-reduced lattice */
	if ((tr = (TheNumeric**)malloc(1 * sizeof(TheNumeric*))) == 0)
		exit (EXIT_FAILURE);
	if ((tr[0] = (TheNumeric*)malloc(1 * sizeof(TheNumeric))) == 0)
		exit (EXIT_FAILURE);
/* perm is the new order in which the vectors in v are added to the lattice
   generated by the preceding vectors */
	if ((perm = (int*)malloc(nv * sizeof(int))) == 0)
		exit (EXIT_FAILURE);
	if ((norm = (TheNumeric*)malloc(nv * sizeof(TheNumeric))) == 0)
		exit (EXIT_FAILURE);
/* bubble sort with respect to the norm of the vectors in v with respect to the
   Gram-matrix F, which is assumed to be positive definite, 
   this should stabilize the LLL-reduction, 
   the bubble sort should be sufficient since the number of vectors here is 
   assumed to be rather small (at most some hundreds) */
	for (i = 0; i < nv; ++i)
	{
		norm[i] = sscp(v[i], Fv[i], dim1);
		if (norm[i] > TheMaxNorm)
		{
			exit (3);
		}
		perm[i] = i;
	}
	i = 0;
	while (i < nv-1)
	{
		if (norm[perm[i]] > norm[perm[i+1]])
		{
			tmp = perm[i];
			perm[i] = perm[i+1];
			perm[i+1] = tmp;
			if (i > 0)
				--i;
		}
		else
			++i;
	}
	free(norm);
/* jump over possible 0-vectors */
	i = 0;
	for (j = 0; j < dim1  &&  v[perm[i]][j] == 0; ++j);
	while (j == dim1  &&  i < nv)
	{
		++i;
		if (i < nv)
			for (j = 0; j < dim1  &&  v[perm[i]][j] == 0; ++j);
	}
	rank = 0;
	while (i < nv)
	{
/* v[perm[i]] is the candidate to enlarge the lattice generated by the 
   base vectors in b */
		vi = v[perm[i]];
		Fvi = Fv[perm[i]];
		for (j = 0; j < rank; ++j)
		{
		  f[j][rank] = sscp((*b)[j], Fvi, dim1);
		  f[rank][j] = f[j][rank];
		}
		f[rank][rank] = sscp(vi, Fvi, dim1);
		if (lll(f, tr, rank+1) > rank)
/* the rank of the lattice generated by v[perm[i]] and the vectors in b is 
   rank+1, i.e. one higher than without v[perm[i]] */
		{
			++rank;
			for (j = 0; j < dim1; ++j)
				(*b)[rank-1][j] = vi[j];
			comtr = com->trans[rank-1];
			for (j = 0; j < nv; ++j)
				comtr[j] = 0;
			comtr[perm[i]] = 1;
/* transform basis and transformation matrix */
			change(*b, rank, tr, dim1);
			change(com->trans, rank, tr, nv);
			if ((*b = (TheNumeric**)realloc(*b, (rank+1) * sizeof(TheNumeric*))) == 0)
				exit (EXIT_FAILURE);
			if (((*b)[rank] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
				exit (EXIT_FAILURE);
			if ((com->trans = (TheNumeric**)realloc(com->trans, (rank+1) * sizeof(TheNumeric*))) == 0)
				exit (EXIT_FAILURE);
			if ((com->trans[rank] = (TheNumeric*)malloc(nv * sizeof(TheNumeric))) == 0)
				exit (EXIT_FAILURE);
			if ((f = (TheNumeric**)realloc(f, (rank+1) * sizeof(TheNumeric*))) == 0)
				exit (EXIT_FAILURE);
			for (j = 0; j < rank; ++j)
			{
				if ((f[j] = (TheNumeric*)realloc(f[j], (rank+1) * sizeof(TheNumeric))) == 0)
					exit (EXIT_FAILURE);
			}
			if ((f[rank] = (TheNumeric*)malloc((rank+1) * sizeof(TheNumeric))) == 0)
				exit (EXIT_FAILURE);
			for (j = 0; j < rank; ++j)
			{
				fj = f[j];
				for (k = 0; k <= j; ++k)
				{
					fj[k] = scp((*b)[j], F, (*b)[k], dim1);
					f[k][j] = fj[k];
				}
			}
			if ((tr = (TheNumeric**)realloc(tr, (rank+1) * sizeof(TheNumeric*))) == 0)
				exit (EXIT_FAILURE);
			for (j = 0; j < rank; ++j)
			{
				if ((tr[j] = (TheNumeric*)realloc(tr[j], (rank+1) * sizeof(TheNumeric))) == 0)
					exit (EXIT_FAILURE);
			}
			if ((tr[rank] = (TheNumeric*)malloc((rank+1) * sizeof(TheNumeric))) == 0)
				exit (EXIT_FAILURE);
		}
		else if (abs(tr[rank][rank]) > 1)
/* v[perm[i]] enlarges the lattice generated by the vectors in b by a finite
   index |tr[rank][rank]| */
		{
			for (j = 0; j < dim1; ++j)
				(*b)[rank][j] = vi[j];
			comtr = com->trans[rank];
			for (j = 0; j < nv; ++j)
				comtr[j] = 0;
			comtr[perm[i]] = 1;
/* transform basis and transformation matrix */
			change(*b, rank+1, tr, dim1);
			change(com->trans, rank+1, tr, nv);
			for (j = 0; j < rank; ++j)
			{
				fj = f[j];
				for (k = 0; k <= j; ++k)
				{
					fj[k] = scp((*b)[j], F, (*b)[k], dim1);
					f[k][j] = fj[k];
				}
			}
		}
		++i;
	}
	com->rank = rank;
	for (i = 0; i < nv; ++i)
		free(Fv[i]);
	free(Fv);
	for (i = 0; i <= rank; ++i)
	{
		free(f[i]);
		free(tr[i]);
	}
	free(f);
	free(tr);
	free(perm);
}

void coef(scpcomb *com, TheNumeric **b, TheNumeric **v, TheNumeric **F, int dim1)
/*****	expresses the vectors in v in the 
      basis b, puts the transformation matrix
      on com->coef, i.e. v = com->coef * b,	
      uses LLL-reduction with respect to F 
      to obtain the coefficients	*****/
{
	int	i, j, nb, nv, sign;
	TheNumeric **Fv, *fnb, **f, *Fvi, *fi, **Fb, **tr, *trnb;
	TheNumeric *comci;
	if ((com->coef = (TheNumeric**)malloc((com->list.n+1) * sizeof(TheNumeric*))) == 0)
		exit (EXIT_FAILURE);
	for (i = 0; i <= com->list.n; ++i)
	{
		if ((com->coef[i] = (TheNumeric*)malloc(com->rank * sizeof(TheNumeric))) == 0)
			exit (EXIT_FAILURE);
	}
	nb = com->rank;
	nv = com->list.n + 1;
/* Fv is the list of products of F with the vectors in v, scalar products
   with respect to F can be calculated as standard scalar products of
   vectors in v with vectors in Fv */
	if ((Fv = (TheNumeric**)malloc(nv * sizeof(TheNumeric*))) == 0)
		exit (EXIT_FAILURE);
	for (i = 0; i < nv; ++i)
	{
	  if ((Fv[i] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
	    exit (EXIT_FAILURE);
	  for (j = 0; j < dim1; ++j)
	    Fv[i][j] = sscp(F[j], v[i], dim1);
	}
/* Fb is the list of products of F with the vectors in b */
	if ((Fb = (TheNumeric**)malloc(nb * sizeof(TheNumeric*))) == 0)
		exit (EXIT_FAILURE);
	for (i = 0; i < nb; ++i)
	{
		if ((Fb[i] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
			exit (EXIT_FAILURE);
		for (j = 0; j < dim1; ++j)
			Fb[i][j] = sscp(F[j], b[i], dim1);
	}
	if ((f = (TheNumeric**)malloc((nb+1) * sizeof(TheNumeric*))) == 0)
		exit (EXIT_FAILURE);
	for (i = 0; i <= nb; ++i)
	{
		if ((f[i] = (TheNumeric*)malloc((nb+1) * sizeof(TheNumeric))) == 0)
			exit (EXIT_FAILURE);
	}
	for (i = 0; i < nb; ++i)
	{
		fi = f[i];
		for (j = 0; j <= i; ++j)
		{
			fi[j] = sscp(b[i], Fb[j], dim1);
			f[j][i] = fi[j];
		}
	}
	if ((tr = (TheNumeric**)malloc((nb+1) * sizeof(TheNumeric*))) == 0)
		exit (EXIT_FAILURE);
	for (i = 0; i <= nb; ++i)
	{
		if ((tr[i] = (TheNumeric*)malloc((nb+1) * sizeof(TheNumeric))) == 0)
			exit (EXIT_FAILURE);
	}
	fnb = f[nb];
	trnb = tr[nb];
	for (i = 0; i < nv; ++i)
	{
		Fvi = Fv[i];
		for (j = 0; j < nb; ++j)
		{
			f[j][nb] = sscp(b[j], Fvi, dim1);
			fnb[j] = f[j][nb];
		}
		fnb[nb] = sscp(v[i], Fvi, dim1);
/* if the vector v[i] is in the lattice generated by the base b, the rank of the
   Gram-matrix f must be nb and the last row of the transformation matrix tr
   expresses the 0-vector, in particular |tr[nb][nb]| must be <=1, since 
   otherwise the vector v[i] would enlarge the lattice be an index 
   |tr[nb][nb]| */
		if (lll(f, tr, nb+1) > nb  ||  abs(trnb[nb]) > 1)
		{
			fprintf(stderr, "Error: vector lies not in lattice\n");
			exit (3);
		}
		else
		{
		  comci = com->coef[i];
		  sign = trnb[nb];
		  for (j = 0; j < nb; ++j)
		    comci[j] = -sign * trnb[j];
		}
	}
	for (i = 0; i < nv; ++i)
		free(Fv[i]);
	free(Fv);
	for (i = 0; i < nb; ++i)
		free(Fb[i]);
	free(Fb);
	for (i = 0; i <= nb; ++i)
	{
		free(f[i]);
		free(tr[i]);
	}
	free(f);
	free(tr);
}

void scpforms(scpcomb *com, TheNumeric **b, invar F)
/*****	com->F[i] is the Gram-matrix of the
      basis b with respect to F.A[i]	*****/
{
	int	i, j, k, dim1, nb;
	TheNumeric **FAi, **Fbi, *comFij;

	dim1 = F.dim1;
	nb = com->rank;
	if ((com->F = (TheNumeric***)malloc(sizeof(TheNumeric**))) == 0)
		exit (EXIT_FAILURE);
/* Fbi is the list of products of F.A[i] with the vectors in b */
	if ((Fbi = (TheNumeric**)malloc(nb * sizeof(TheNumeric*))) == 0)
		exit (EXIT_FAILURE);
	for (j = 0; j < nb; ++j)
	{
	  if ((Fbi[j] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
	    exit (EXIT_FAILURE);
	}
	for (i = 0; i < 1; ++i)
	{
	  if ((com->F[i] = (TheNumeric**)malloc(nb * sizeof(TheNumeric*))) == 0)
	    exit (EXIT_FAILURE);
	  FAi = F.A[i];
	  for (j = 0; j < nb; ++j)
	    {
	      for (k = 0; k < dim1; ++k)
		Fbi[j][k] = sscp(FAi[k], b[j], dim1);
	    }
	  for (j = 0; j < nb; ++j)
	    {
	      if ((com->F[i][j] = (TheNumeric*)malloc(nb * sizeof(TheNumeric))) == 0)
		exit (EXIT_FAILURE);
	      comFij = com->F[i][j];
	      for (k = 0; k < nb; ++k)
		comFij[k] = sscp(b[j], Fbi[k], dim1);
	    }
	}
	for (j = 0; j < nb; ++j)
	  free(Fbi[j]);
	free(Fbi);
}

int comp(TheNumeric *x, TheNumeric *y, int n)
/*****	compares the 1xn-vectors x and y lexicographically
      returns 1 if x > y, 0 if x = y, -1 if x < y	*****/
{
	int	i;
  
	for (i = 0; i < n  &&  x[i] == y[i]; ++i);
	if (i == n)
		return(0);
	else if (x[i] > y[i])
		return(1);
	else
		return(-1);
}

int numberof(TheNumeric *vec, veclist V)
/*****	searches for the vector vec in sorted 
      list v := V.v between v[1] and v[V.n],
      where the first non-zero entry in v[i] is > 0,
      returns i > 0 if vec = v[i],
      returns -i < 0 if -vec = v[i],
      if vector is not found it returns
      V.n+i if v[i-1] < vec < v[i] or
      -(V.n+i) if v[i-1] < -vec < v[i]
      if the return value is negative, vec is
      replaced by -vec	*****/
{
	int	i, sign, dim1, low, high, search, cmp;

	sign = 1;
	dim1 = V.dim1;
	low = 1;
	high = V.n;
	for (i = 0; i < dim1  &&  vec[i] == 0; ++i);
	if (i < dim1  &&  vec[i] < 0)
	{
		sign = -1;
		for (i = 0; i < dim1; ++i)
			vec[i] *= -1;
	}
	while (low <= high)
	{
		search = low + (high-low)/2;
		cmp = comp(vec, V.v[search], dim1);
		if (cmp == 1)
/* the vector is in the upper half */
			low = search + 1;
		else if (cmp == -1)
/* the vector is in the lower half */
			high = search - 1;
		else 
/* the vector is found */
			return(sign * search);
	}
/* if low > high the vector was not found */
	return(sign * (V.n+low));
}

void sortvecs(veclist *V)
/*****	sorts the V->n vectors v[1]...v[V->n] and 
      deletes doublets, V->v is changed !!!	*****/
{
	int	i, j;

/* sort the vectors */
	quicksort(V->v, 1, V->n, V->dim1);
/* now delete doublets */
	j = 1;
	for (i = 1; i+j <= V->n; ++i)
	{
		while (i+j <= V->n  &&  comp(V->v[i], V->v[i+j], V->dim1) == 0)
		{
			free(V->v[i+j]);
			++j;
		}
		if (i+j <= V->n)
		{
			V->v[i+1] = V->v[i+j];
			if (comp(V->v[i], V->v[i+1], V->dim1) == 1)
			{
				fprintf(stderr, "Error: v[%d] > v[%d]\n", i, i+1);
				exit(2);
			}
		}
	}
	V->n -= j-1;
}

void quicksort(TheNumeric **v, int inf, int sup, int dim1)
/*****	standard quicksort	*****/
{
  int	low, med, high;
  TheNumeric *tmp;
  if(inf+1 < sup)
    {
      /* interchange v[inf] and v[med] to avoid worst case for pre-sorted lists */
      med = (inf+1+sup)/2;
      tmp = v[inf];
      v[inf] = v[med];
      v[med] = tmp;
      low = inf+1;
      high = sup;
      while(low < high)
	{
	  while(comp(v[inf], v[low], dim1) >= 0  &&  low < sup)
	    ++low;
	  while(comp(v[inf], v[high], dim1) <= 0  &&  high > inf)
	    --high;
	  if(low < high)
	    {
	      tmp = v[high];
	      v[high] = v[low];
	      v[low] = tmp;
	    }
	}
      if(inf != high)
	{
	  tmp = v[high];
	  v[high] = v[inf];
	  v[inf] = tmp;
	}
      quicksort(v, inf, high-1, dim1);
      quicksort(v, high+1, sup, dim1);
    }
  if(inf+1 == sup)
    {
      if(comp(v[inf], v[sup], dim1) == 1)
	{
	  tmp = v[inf];
	  v[inf] = v[sup];
	  v[sup] = tmp;
	}
    }
}

/* main function of ISOM */

int isom(double* form1, double* form2)
{
//	FILE		*infile;
//	char		*filename;
	invar		F, FF;
	veclist		V, norm;
	fpstruct	fp;
	int		dim1;
	int		nG;
	int		i, j, k, nV1;
	TheNumeric      max;
	TheNumeric      **FAi, *Vvi, *Vvj, **Fvi, *Fvij;
	TheNumeric      *vec, ***G;
	TheNumeric      TheMaxNorm, TheMaxEntry;
	TheMaxNorm=GetMAXNORM();
	TheMaxEntry=GetMAXENTRY();

  double** matrix1;
  double** matrix2;

  doubleMakeSquareMatrix(dim,&matrix1);
  doubleMakeSquareMatrix(dim,&matrix2);

  double det1;
  double det2;

  for(int i=0;i<dim;i++)
	for(int j=0;j<=i;j++){
		matrix1[i][j]=form1[i*(i+1)/2+j];
		matrix1[j][i]=form1[i*(i+1)/2+j];
		matrix2[i][j]=form2[i*(i+1)/2+j];
		matrix2[j][i]=form2[i*(i+1)/2+j];
	}

  determinant(dim,matrix1,&det1);
  if( ((fmod(det1,1.0))>0.01)&&((fmod(det1,1.0))<0.99)) printf("ALERT! det1 is non-integer.\n");


  determinant(dim,matrix2,&det2);

  if( ((fmod(det2,1.0))>0.01)&&((fmod(det2,1.0))<0.99)) printf("ALERT! det1 is non-integer.\n");

  doubleDestroySquareMatrix(dim,&matrix1);
  doubleDestroySquareMatrix(dim,&matrix2);

  free(matrix1);
  free(matrix2);

  if(fabs(det1-det2)>0.1) {
	return 0;
  }

	if ((F.A = (TheNumeric***)malloc(sizeof(TheNumeric**))) == 0)
		exit (1);
/* read the invariant forms of the first lattice */

	F.dim1=dim;

	dim1 = F.dim1;

	if ((F.A[0] = (TheNumeric**)malloc(dim1 * sizeof(TheNumeric*))) == 0)
	  exit (1);
	for (i = 0; i < dim1; ++i)
	{
		if (((F.A[0])[i] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
			exit (1);
	}

    for(i=0;i<dim1;++i)
		for(j=0;j<=i;++j)
		{
			F.A[0][i][j]=form1[i*(i+1)/2+j];
			F.A[0][j][i]=form1[i*(i+1)/2+j];
		}

	V.dim1 = dim1;
	V.len = dim1 + 1;

/* the short vectors are calculated (which is the default) */
		max = F.A[0][0][0];
		for (i = 1; i < dim1; ++i)
		{
			if (F.A[0][i][i] > max)
				max = F.A[0][i][i];
		}
/* the vectors with length up to the maximum of the diagonal of the first
   form (which is assumed to be positive definite) are required */
		shortvec(F.A[0], max, &V);

/* the first nonzero entry of each vector is made positive and the maximal 
   entry in the vectors is determined */
	max = 1;
	for (i = 1; i <= V.n; ++i)
	{
		Vvi = V.v[i];
		for (j = 0; j < dim1  &&  Vvi[j] == 0; ++j);
		if (j < dim1  &&  Vvi[j] < 0)
		{
			for (k = j; k < dim1; ++k)
				Vvi[k] *= -1;
		}
		for (k = j; k < dim1; ++k)
		{
			if (abs(Vvi[k]) > max)
				max = abs(Vvi[k]);
		}
	}
	if (max > TheMaxEntry)
/* there is an entry, which is bigger than MAXENTRY, which might cause overflow,
   since the program doesn't use long integer arithmetic */
	{
	  fprintf(stderr, "Error: Found entry ");
	  single2Print(stderr, max);
	  fprintf(stderr, " in short vectors of the first lattice.\nTo avoid overflow, the entries should not exceed ");
	  single2Print(stderr, TheMaxEntry);
	  fprintf(stderr, ".\n");
	  exit (3);
	}
/* sort the vectors and delete doublets */
	sortvecs(&V);
/* the norm-vector (i.e. the vector of the norms with respect to the different
   invariant forms) of each vector must be equal to the norm-vector of one
   of the vectors from the standard-base */
	if ((norm.v = (TheNumeric**)malloc((dim1+1) * sizeof(TheNumeric*))) == 0)
		exit (1);
	norm.n = dim1;
	norm.dim1 = 1;
	norm.len = 1;
	for (i = 1; i <= norm.n; ++i)
	{
/* norm.v[i] is the norm combination of the (i-1)-th base-vector */
	  if ((norm.v[i] = (TheNumeric*)malloc(norm.dim1 * sizeof(TheNumeric))) == 0)
	    exit (1);
	  for (k = 0; k < norm.dim1; ++k)
	    norm.v[i][k] = F.A[k][i-1][i-1];
	}
	sortvecs(&norm);
/* delete those vectors, which can not be the image of any vector of the
   standard-base */
	checkvecs(&V, F, norm);
/* F.v[i][j] is the transposed of the product of the Gram-matrix F.A[i] with the
   transposed of the vector v[j], hence the scalar product of v[j] and v[k] with
   respect to the Gram-matrix F.A[i] can be computed as standard scalar product
   of v[j] and F.v[i][k] */
	if ((F.v = (TheNumeric***)malloc(sizeof(TheNumeric**))) == 0)
		exit (1);
/* the product of the maximal entry in the short vectors with the maximal entry
   in F.v[i] should not exceed MAXNORM to avoid overflow */
	max = TheMaxNorm / max;
	for (i = 0; i < 1; ++i)
	{
		FAi = F.A[i];
		if ((F.v[i] = (TheNumeric**)malloc((V.n+1) * sizeof(TheNumeric*))) == 0)
			exit (1);
		Fvi = F.v[i];
		for (j = 1; j <= V.n; ++j)
		{
			Vvj = V.v[j];
			if ((Fvi[j] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
				exit (1);
			Fvij = Fvi[j];
			for (k = 0; k < dim1; ++k)
			{
				Fvij[k] = sscp(FAi[k], Vvj, dim1);
				if (abs(Fvij[k]) > max)
/* some entry in F.v[i] is too large */
				{
				  fprintf(stderr, "Error: Found entry ");
				  single2Print(stderr, Fvij[k]);
				  fprintf(stderr, " in F.v.\nTo avoid overflow, the entries should not exceed ");
				  single2Print(stderr, max);
				  fprintf(stderr, ".\n");
				  exit (3);
				}
			}
		}
	}
/* fp.per is the order in which the images of the base vectors are chosen */
	if ((fp.per = (int*)malloc(dim1 * sizeof(int))) == 0)
		exit (1);
/* fp.e[i] is the index in V.v of the i-th vector of the standard-base in the 
   new order */
	if ((fp.e = (int*)malloc(dim1 * sizeof(int))) == 0)
		exit (1);
/* fp.diag is the diagonal of the fingerprint in the new order */
	if ((fp.diag = (int*)malloc(dim1 * sizeof(int))) == 0)
		exit (1);
/* compute the fingerprint */
	fingerprint(&fp, F, V);
/* get the standard basis in the new order */
	if ((vec = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
		exit (1);
	for (i = 0; i < dim1; ++i)
	{
		for (j = 0; j < dim1; ++j)
			vec[j] = 0;
		vec[fp.per[i]] = 1;
		fp.e[i] = numberof(vec, V);
		if (abs(fp.e[i]) > V.n)
		{
			fprintf(stderr, "Error: standard basis vector nr. %d not found\n", i);
			exit (2);
		}
	}
	free(vec);

/* the vectors of the first lattice are no longer required, only their number */
	nV1 = V.n;
	for (i = 0; i <= V.n; ++i)
		free(V.v[i]);
	free(V.v);
	for (i = 0; i < 1; ++i)
	{
		for (j = 1; j <= V.n; ++j)
			free(F.v[i][j]);
		free(F.v[i]);
	}
	free(F.v);



//	FF.n=1;

	if ((FF.A = (TheNumeric***)malloc(sizeof(TheNumeric**))) == 0)
		exit (1);
/* read the invariant forms of the first lattice */

	FF.dim1=dim1;

	if ((FF.A[0] = (TheNumeric**)malloc(dim1 * sizeof(TheNumeric*))) == 0)
	  exit (1);
	for (i = 0; i < dim1; ++i)
	{
		if (((FF.A[0])[i] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
			exit (1);
	}

    for(i=0;i<dim1;++i)
		for(j=0;j<=i;++j)
		{
//			fscanf(infile,"%d",&temp);
//			temp=atoi(argv[dim1*(dim1+1)/2+i*(i+1)/2+j+2]);
			FF.A[0][i][j]=form2[i*(i+1)/2+j];
			FF.A[0][j][i]=form2[i*(i+1)/2+j];
		}

/* the short vectors of the second lattice are calculated (which is the default) */
	
		max = F.A[0][0][0];
		for (i = 1; i < dim1; ++i)
		{
			if (F.A[0][i][i] > max)
				max = F.A[0][i][i];
		}
/* the vectors with length up to the maximum of the diagonal of the first
   form of the first lattice are required */
		shortvec(FF.A[0], max, &V);
	
/* the first nonzero entry of each vector is made positive and the maximal
   entry of the vectors is determined */
	max = 1;
	for (i = 1; i <= V.n; ++i)
	{
		Vvi = V.v[i];
		for (j = 0; j < dim1  &&  Vvi[j] == 0; ++j);
		if (j < dim1  &&  Vvi[j] < 0)
		{
			for (k = j; k < dim1; ++k)
				Vvi[k] *= -1;
		}
		for (k = j; k < dim1; ++k)
		{
			if (abs(Vvi[k]) > max)
				max = abs(Vvi[k]);
		}
	}
	if (max > TheMaxEntry)
/* there is an entry, which is bigger than MAXENTRY, which might cause overflow,
   since the program doesn't use long integer arithmetic */
	{
	  fprintf(stderr, "Error: Found entry ");
	  single2Print(stderr, max);
	  fprintf(stderr, " in short vectors of the second lattice.\nTo avoid overflow, the entries should not exceed ");
	  single2Print(stderr, TheMaxEntry);
	  fprintf(stderr, ".\n");
	  exit (3);
	}
/* V.prime is a prime p, such that the entries of the short vectors remain 
   unchanged under reduction mod p in symmetric form, i.e. -p/2 < x <= p/2 */
	for (V.prime = 2*max + 1; isprime(V.prime) == 0; ++V.prime);
	k = F.A[0][0][0];
	for (i = 1; i < dim1; ++i)
	{
		if (F.A[0][i][i] > k)
			k = F.A[0][i][i];
	}
	for (i = 0; i < dim1; ++i)
	{
		if (FF.A[0][i][i] > k)
			V.prime = DEF_PRIME;
	}
/* sort the vectors and delete doublets */
	sortvecs(&V);
/* the norm-vector (i.e. the vector of the norms with respect to the different
   invariant forms) of each vector must be equal to the norm-vector of one
   of the vectors from the standard-base of the first lattice */
	checkvecs(&V, FF, norm);
	for (i = 1; i <= norm.n; ++i)
		free(norm.v[i]);
	free(norm.v);
	if (V.n != nV1)
	{

// Free memory
		for(i=0; i<dim1; i++){
			free((F.A[0])[i]);
			free((FF.A[0])[i]);
		}
		free(F.A[0]);
		free(FF.A[0]);

		free(F.A);
		free(FF.A);
			
		for (i = 0; i <= V.n; ++i)
			free(V.v[i]);
		free(V.v);
	
		free(fp.e);
		free(fp.diag);
		free(fp.per);

		return 0;
		printf("The lattices are not isomorphic !\nDifferent numbers of short vectors (%d|%d)\n", nV1, V.n);
		exit (0);
	}
/* FF.v[i][j] is the transposed of the product of the Gram-matrix FF.A[i] 
   with the transposed of the vector V.v[j], which now is a short vector of
   the second lattice */
	if ((FF.v = (TheNumeric***)malloc(sizeof(TheNumeric**))) == 0)
		exit (1);
/* the product of the maximal entry in the short vectors with the maximal entry
   in FF.v[i] should not exceed MAXNORM to avoid overflow */
	max = TheMaxNorm / max;
	for (i = 0; i < 1; ++i)
	{
		FAi = FF.A[i];
		if ((FF.v[i] = (TheNumeric**)malloc((V.n+1) * sizeof(TheNumeric*))) == 0)
			exit (1);
		Fvi = FF.v[i];
		for (j = 1; j <= V.n; ++j)
		{
			Vvj = V.v[j];
			if ((Fvi[j] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
				exit (1);
			Fvij = Fvi[j];
			for (k = 0; k < dim1; ++k)
			{
				Fvij[k] = sscp(FAi[k], Vvj, dim1);
				if (abs(Fvij[k]) > max)
/* some entry in FF.v[i] is too large */
				{
				  fprintf(stderr, "Error: Found entry ");
				  single2Print(stderr, Fvij[k]);
				  fprintf(stderr, " in FF.v.\nTo avoid overflow, the entries should not exceed ");
				  single2Print(stderr, max);
				  fprintf(stderr, ".\n");
				  exit (3);
				}
			}
		}
	}
/* G are the automorphisms of the second lattice */
	if ((G = (TheNumeric***)malloc(1 * sizeof(TheNumeric**))) == 0)
		exit (1);
/* nG is the number of automorphisms of the second lattice */
	nG = 0;

	if (nG == 0)
/* if no automorphisms are known at least take -Id */
	{
		if ((G[0] = (TheNumeric**)malloc(dim1 * sizeof(TheNumeric*))) == 0)
			exit (1);
		for (i = 0; i < dim1; ++i)
		{
			if ((G[0][i] = (TheNumeric*)malloc(dim1 * sizeof(TheNumeric))) == 0)
				exit (1);
			for (j = 0; j < dim1; ++j)
				G[0][i][j] = 0;
			G[0][i][i] = -1;
		}
		nG = 1;
	}

/* now search for an isometry */
	int answer;
	answer=isometry(V, F, FF, fp, G, nG);

// Free memory

	for(i=0; i<dim1; i++){
		free((F.A[0])[i]);
		free((FF.A[0])[i]);
	}
	free(F.A[0]);
	free(FF.A[0]);
		
	for (i = 0; i <= V.n; ++i)
		free(V.v[i]);
	free(V.v);

//	for (j = 1; j <= V.n; ++j)
//		free(Fvi[j]);


	for (i = 0; i < 1; ++i)
	{
		for (j = 1; j <= V.n; ++j)
			free(FF.v[i][j]);
		free(FF.v[i]);
	}
	free(FF.v);

	free(fp.e);
	free(fp.diag);
	free(fp.per);

	free(F.A);
	free(FF.A);


//	TheNumeric      **FAi, *Vvi, *Vvj, **Fvi, *Fvij;
//	TheNumeric      **sumveclist, **sumvecbase, *vec, ***G;
 

	return answer;
}
