// DeloneTriangulation.h

#ifndef _DELONE_TRIANGULATION_H
#define _DELONE_TRIANGULATION_H

#include <iostream>
#include <vector>
#include <ext/hash_set>
#include <set>
#include <list>
#include <NTL/mat_ZZ.h>

#include "Voronoi.h"
#include "Simplex.h"
#include "SimplexContainer.h"
#include "Facet.h"
#include "Regulator.h"

#include "Extreme.h"
//#include "QForm.h"

//#include "ISOM.h"

typedef __gnu_cxx::hash_set<Simplex*, SimplexHash, SimplexEq> hash_simplices;

typedef struct {
  double* q;    // entries of the quadratic form
  int* ranks;	// number of extreme forms of certain ranks
//  int codim;	// codimension in the Secondary cone
  int* marker;
  int parent;
  int codim;
} QForm;

class DeloneTriangulation {
 public:
  DeloneTriangulation();
  DeloneTriangulation(const DeloneTriangulation& Del);
  ~DeloneTriangulation();

  DeloneTriangulation& operator=(const DeloneTriangulation& Del);

  bool is_equivalent(DeloneTriangulation& Del);

  int compute_no_neighbours();
  DeloneTriangulation compute_nth_neighbour(const int n);
  // before calling compute_nth_neighbour you have to call compute_no_neighbours

  void compute_lower_bound(NTL::ZZ& invariant, double& lower_bound);

  void compute_invariant(NTL::ZZ& invariant); // A.G. trying to remove computing lower bound.

  void read_COOP_file(FILE* coop_file);
  void write_COOP_file(FILE* coop_file) const;

  void write_form_file(QForm form, FILE* form_file);

  int no_simplices(); // A.G.

  double volume_sq_sum(); // A.G.

  int compute_max_volume(); // A.G.

  Extreme compute_nth_extreme(const int n);
  int compute_no_extreme();

  void subset(const int *marker, QForm* res);

// codim - codimension of faces what we are looking for. If marker==(1..1) then codim=1.
  int subset_facets(const int codim, const int* marker, int** res_marker);

  friend DeloneTriangulation principal_domain();
  friend DeloneTriangulation barnes_trenerry_domain();

 private:
  hash_simplices simplices;
  int no_neighbours;
  int no_extreme;
  
  NTL::ZZ invar; // A.G.

  int max_volume; // A.G.

  std::set<Facet> facets; // facets of Delone triangulations
  std::set<Regulator> regulators;
  std::set<Extreme> extreme_forms; // A.G. Set of extreme forms
  std::vector<std::set<Regulator>::iterator> irredundants; // facets of secondary cone

  void make_neighbours();
  void destroy_neighbours();
  void make_facets_and_regulators();
  void destroy_facets_and_regulators();
  DeloneTriangulation do_flip(std::list<Facet*> flip_facets) const;

  void make_extreme();
  int* extreme_ranks;

  static SimplexContainer all_simplices;
};

DeloneTriangulation principal_domain();

DeloneTriangulation barnes_trenerry_domain();

/* Adding ISOM */

int isom(double *form1, double *form2);

#endif

#ifdef ARITH64
typedef long long TheNumeric;
#else
typedef int TheNumeric;
#endif



/* the maximal entry in the short vectors should not exceed MAXENTRY to
   avoid overflow */
/* it is checked that the product of the maximal entry in the short vectors
   and the maximal entry in the products of Gram-matrices with short vectors
   does not exceed MAXNORM, since the program works just with int's */
#define	STRLEN		80
/* since the LLL-reduction works with a floating-point model, there might be
   some rounding error, which should not reach EPS */
#define	EPS		0.001
/* the constant for the LLL-condition, which should be < 1, but 0.75 is the
   standard value, a higher value might yield a slightly better result but
   increases the running time */
#define	LLL_CONST	0.75
#define	DEF_PRIME	16001

#ifndef SEEK_SET		/* some UNIX-systems do not define SEEK_SET */
#	define SEEK_SET	0	/* Set file pointer to "offset" */
#endif

/* structure to hold the generators of the group according to the stabilizer
   chain */
typedef struct {
	int	dim1;
	int	*ord;
	int	*ng;
	int	*nsg;
	TheNumeric	****g;
	} group;

/* structure for a list of vectors */
typedef struct {
  int	dim1;
  int	len;
  TheNumeric	prime;
  int	n;
  TheNumeric	**v;
} veclist;

/* structure for the invariant forms and their products with the list of
   short vectors */
typedef struct {
	int	dim1;
//	int	n;
	TheNumeric	***A;
	TheNumeric	***v;
	} invar;

/* structure for the diagonal of the fingerprint, the new order and the indices
   of the standard-base in the list of short vectors */
typedef struct {
	int	*diag;
	int	*per;
	int	*e;
	} fpstruct;

/* structure for the scalar product combinations, the transformation matrices
   between the vector sums and that lattice bases and the scalar products of
   the lattice bases */

 typedef struct {
	veclist	list;
	int	rank;
	TheNumeric	**trans;
	TheNumeric	**coef;
	TheNumeric	***F;
	} scpcomb; 


/* functions in iotools.c */
void single2Print(FILE *f, TheNumeric TheNb);
// void getflags(flagstruct *fl, int argc, char *argv[]);
//int getnumber(FILE *infile);
//int getmat(FILE *infile, TheNumeric ***A);
//void getvecs(FILE *infile, veclist *V);
void putform(TheNumeric **A, int n);
// void putgens(group G, flagstruct flags);
TheNumeric GetMAXNORM();
TheNumeric GetMAXENTRY();
// void putord(group G, flagstruct flags);
void putiso(TheNumeric **X, int dim1);

/* functions in isotools.c */
int isocand(int *CI, int I, int *x, veclist V, invar F, invar FF, fpstruct fp);
int isometry(veclist V, invar F, invar FF, fpstruct fp, TheNumeric ***G, int nG);
int iso(int step, int *x, int **C, veclist V, invar F, invar FF, fpstruct fp, TheNumeric ***G, int nG);
int isostab(TheNumeric ****H, int pt, TheNumeric ***G, int nG, veclist V, int Maxfail);

/* functions in lattools.c */
int lll(TheNumeric **F, TheNumeric **T, int dim1);
void initialize(int i, double **mu, double *B, TheNumeric **gram);
int red(int *m, int *l, TheNumeric **gram, TheNumeric **T, double **mu, double *B, int dim1, int *rank);
void check(double *B, int *m, int *l, TheNumeric **gram, TheNumeric **T, double **mu, int dim1, int *rank);
void decrease(int *m, int *l, TheNumeric **gram, double **mu, double *B, int *rank);
void interchange(int *m, int *l, TheNumeric **gram, TheNumeric **T, double **mu, double *B, int dim1);
TheNumeric MyRound(double x);
void shortvec(TheNumeric **F, int max, veclist *V);
void shrt(int k, double norm, int max, veclist *V, TheNumeric *vec, double **mu, TheNumeric **T, double *B, int *con);
void change(TheNumeric **v, int nv, TheNumeric **T, int n);

/* functions in mattools.c */
void vecmatmul(TheNumeric *x, TheNumeric **A, int n, TheNumeric *y);
void matmul(TheNumeric **A, TheNumeric **B, int n, TheNumeric **C);
TheNumeric scp(TheNumeric *x, TheNumeric **F, TheNumeric *y, int n);
TheNumeric sscp(TheNumeric *x, TheNumeric *y, int n);
void psolve(TheNumeric **X, TheNumeric **A, TheNumeric **B, int n, TheNumeric p);
void pgauss(int r, TheNumeric **A, TheNumeric **B, int n, TheNumeric p);
int isprime(TheNumeric n);

/* functions in orbtools.c */
int operate(int nr, TheNumeric **A, veclist V);
int orbit(int	*pt, int npt, TheNumeric ***G, int nG, veclist V, int **orb);
int orbitlen(int pt, int orblen, TheNumeric ***G, int nG, veclist V);
int delete_new(int *orb1, int l1, int *orb2, int l2);
void stab(int I, group *G, fpstruct fp, veclist V);
void matgen(int *x, TheNumeric **X, int dim1, int *per, TheNumeric **v);
void stabil(TheNumeric **S, int *x1, int *x2, int *per, TheNumeric **G, veclist V);

/* functions in preproc.c */
void    checkvecs(veclist *V, invar F, veclist norm);
int     checkgen(TheNumeric **g, invar F);
void    fingerprint(fpstruct *fp, invar F, veclist V);
int     possible(invar F, TheNumeric ***Ftr, veclist V, int *per, int I, int J);
void    scpvector(TheNumeric *scpvec, TheNumeric *w, int *b, int I, int dep, invar F);
void    scpvecs(veclist *list, TheNumeric ***vec, int I, int *b, int dep, veclist V, invar F);
void    base(scpcomb *com, TheNumeric ***b, TheNumeric **v, TheNumeric **F, int dim1);
void    coef(scpcomb *com, TheNumeric **b, TheNumeric **v, TheNumeric **F, int dim1);
void    scpforms(scpcomb *com, TheNumeric **b, invar F);

/* functions in sorttools.c */
int comp(TheNumeric *x, TheNumeric *y, int n);
int numberof(TheNumeric *vec, veclist V);
void sortvecs(veclist *V);
void quicksort(TheNumeric **v, int inf, int sup, int dim1);
