// Simplex.cc

#include <algorithm>
#include <functional>
#include <iostream>
#include <sstream>
#include <cmath>
#include <NTL/mat_ZZ.h>

#include "Voronoi.h"
#include "Simplex.h"
#include "la_support.h"

#include <cstdio>

//
//
// 

// ok, this hash function is somehow stupid, any better suggestions?

size_t compute_hash(double** vertices) 
{
  int hash_no = 0;
  for (int i = 1; i < dim+1; i++)
    for (int j = 0; j < dim; j++)
      hash_no = hash_no + (j*j*j+i*i*i+i*i+j*j)*int(vertices[i][j]);
  return hash_no;
}

inline bool compare_vertices(const double* v, const double* w) 
{  
  int i = 0;
  while (v[i] == w[i] && i < dim) i++;
  return v[i] < w[i];
}

void normalize_vertices(double** v)
{
  std::sort(&(v[0]), &(v[dim+1]), compare_vertices);

  double w;
  for (int j = 0; j < dim; j++) {
    w = v[0][j];
    for (int i = 1; i < dim+1; i++)
      v[i][j] -= w;
  }

  std::fill(&(v[0][0]), &(v[0][dim]), 0.0);
}

//
//
//

Simplex::Simplex(double** v)
{
  doubleMakeSquareMatrix(dim+1, &vertices);
  doubleCopySquareMatrix(dim+1, v, vertices);
  normalize_vertices(vertices);
  hash_no = compute_hash(vertices);
  volume=0;
//  volume=compute_volume();
}

Simplex::Simplex(const Simplex& L)
{
  doubleMakeSquareMatrix(dim+1, &vertices);
  doubleCopySquareMatrix(dim+1, L.vertices, vertices);
  hash_no = L.hash_no;
  volume=L.volume;
}

Simplex& Simplex::operator=(const Simplex& L)
{
  if (this != &L) {
    doubleCopySquareMatrix(dim+1, L.vertices, vertices);
    hash_no = L.hash_no;
    volume=L.volume;
  }
  return *this;
}

Simplex::~Simplex()
{
  doubleDestroySquareMatrix(dim+1, &vertices);
}

Simplex Simplex::compute_negative() const
{
  double** neg_vertices;
  doubleMakeSquareMatrix(dim+1, &neg_vertices);
  doubleCopySquareMatrix(dim+1, vertices, neg_vertices);
  for (int i = 1; i < dim+1; i++)
    for(int j = 0; j < dim; j++)
      neg_vertices[i][j] = -neg_vertices[i][j];
  Simplex L(neg_vertices);
  doubleDestroySquareMatrix(dim+1, &neg_vertices);
  return L;
}

//
// 
//

void Simplex::get_nth_facet(const int n, double** facet) const
{
  doubleCopySquareMatrix(dim+1, vertices, facet);
  std::rotate(&(facet[n]), &(facet[n+1]), &(facet[dim+1]));
  if (n == 0) {
    doubleCopyVector(dim+1, facet[0], facet[dim]);
    doubleMultiplyVector(dim, -1.0, facet[dim], facet[dim]);
    for (int i = 0; i < dim; i++)
      doubleAddVector(dim, 1.0, facet[dim], facet[i]);
  }
}

//
//
//

void Simplex::compute_inverse(double** inv_vertices, double& inv_det) const
{  
  NTL::ZZ det;
  NTL::mat_ZZ mat;
  mat.SetDims(dim,dim);
  for (int i = 0; i < dim; i++)
    for (int j = 0; j < dim; j++)
      mat[i][j] = NTL::to_ZZ(vertices[i+1][j]);
  NTL::mat_ZZ inv_mat;
  inv_mat.SetDims(dim,dim);
  inv(det, inv_mat, mat, true);
  inv_det = NTL::to_double(det);

  for (int i = 0; i < dim; i++)
    for (int j = 0; j < dim; j++)
      inv_vertices[i][j] = NTL::to_double(inv_mat[i][j]);
}

bool Simplex::find_linear_map(const Simplex& L, const int n, const int* pi, 
			      double** inv_vertices, double inv_det, double** linear_map) const
{
  // Warning! row vectors...
  double** Ln_vertices;
  doubleMakeSquareMatrix(dim, &Ln_vertices);
  int j = 0;
  for (int i = 0; i < dim+1; i++) {
    if (i != n) {
      for (int k = 0; k < dim; k++) {
	int p = pi[j];
	Ln_vertices[p][k] = L.vertices[i][k] - L.vertices[n][k];
      }
      j++;
    }
  }
  doubleMultiplySquareMatrix(dim, inv_vertices, Ln_vertices, linear_map);

  for (int i = 0; i < dim; i++)
    for (int j = 0; j < dim; j++) {
      if (int(floor(linear_map[i][j] + 0.5)) % int(floor(inv_det + 0.5)) != 0)
	return false;
      linear_map[i][j] = floor(linear_map[i][j]/inv_det + 0.5);
    }

  doubleDestroySquareMatrix(dim, &Ln_vertices);
  return true;
}

Simplex Simplex::perform_linear_map(double** linear_map) const
{
  // Warning! row vectors...
  double** tmp_vertices;
  doubleMakeSquareMatrix(dim+1, &tmp_vertices);
  for (int i = 0; i < dim+1; i++) 
    tmp_vertices[i][dim] = 1.0;
  doubleMultiplySquareMatrix(dim, &(vertices[1]), linear_map, &(tmp_vertices[1]));
  Simplex L(tmp_vertices);
  doubleDestroySquareMatrix(dim+1, &tmp_vertices);
  return L;
}

//
//
//

// strange function which is needed for the computation of the lower bound
// of the covering density of a secondary cone

void Simplex::inc_edge_matrix(double **edge_matrix) const
{
  
  for (int i = 0; i < dim+1; i++)
    for (int j = i+1; j < dim+1; j++)
      for (int k = 0; k < dim; k++)
	for (int l = 0; l < dim; l++)
	  edge_matrix[k][l] += 
	    (vertices[i][k] - vertices[j][k]) * 
	    (vertices[i][l] - vertices[j][l]);
}

//
//
//

void Simplex::compute_circumsphere_matrix(const int i, const int j, double *F) const
{
  F[0] = 0.0;
  if (i == j) {
    for (int k = 1; k < dim+1; k++)
      F[k] = vertices[k][i] * vertices[k][i];
    for (int k = 0; k < dim; k++)
      for (int l = k; l < dim;l++)
	F[(dim+1) + k*dim+l-k*(k+1)/2] = vertices[k+1][i]*vertices[l+1][i];
  }
  else {
    for (int k = 1; k < dim+1; k++)
      F[k] = 2.0 * vertices[k][i] * vertices[k][j];
    for (int k = 0; k < dim; k++)
      for (int l = k; l < dim;l++)
	F[(dim+1) + k*dim+l-k*(k+1)/2] 
	  = vertices[k+1][i]*vertices[l+1][j] + vertices[l+1][i]*vertices[k+1][j];
  }
}

//
//
//

size_t Simplex::compute_hash_no() const
{
  return hash_no;
}

//
//
//

bool Simplex::operator==(const Simplex& L) const
{
  return !(*this < L) && !(L < *this);
}

bool Simplex::operator<(const Simplex& L) const
{
  if (hash_no == L.hash_no) {
    for (int i = 1; i < dim+1; i++)
      for (int j = 0; j < dim; j++) {
	double sign = vertices[i][j] - L.vertices[i][j];
	if (sign < -0.5) return true;
	if (sign > 0.5) return false;
      }
    return false;
  }
  return hash_no < L.hash_no;
}

// this is definitely not state of the art

void Simplex::print_me(FILE* f) const
{
  for (int i = 0; i < dim; i++) {
    for (int j = 0; j < dim; j++)
      fprintf(f, "%d ",  int(vertices[i+1][j]));
    fprintf(f, "\n");
    }
  fprintf(f, "\n");
}

// A.G.

int Simplex::compute_volume()
{  
  if(volume!=0)
    return volume;

  NTL::ZZ det;
  NTL::mat_ZZ mat;
  mat.SetDims(dim,dim);
  for (int i = 0; i < dim; i++)
    for (int j = 0; j < dim; j++)
      mat[i][j] = NTL::to_ZZ(vertices[i+1][j]);
  det=determinant(mat,true);
  volume=fabs(NTL::to_int(det));

  return volume;
}

